#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest
from ai_flow.workflow.workflow_config import WorkflowConfig
from ai_flow.workflow.job_config import JobConfig
from ai_flow.context.job_context import job_config, current_job_name
from ai_flow.context.workflow_config_loader import current_workflow_config


class TestJobContext(unittest.TestCase):

    def test_job_context(self):
        workflow_config = current_workflow_config()
        workflow_config.job_configs['job_1'] = JobConfig(job_name='job_1')
        with job_config('job_1') as jc:
            self.assertEqual(jc.job_name, current_job_name())


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import time

from airflow.executors.scheduling_action import SchedulingAction
from airflow.utils import dates
from enum import Enum
from notification_service.base_notification import BaseEvent
import json

EXECUTION_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
SCHEDULER_NAMESPACE = 'scheduler'
UNREACHED_EVENT = 'UNREACHED_EVENT'


# set task trigger only by scheduler inner events
class UnreachedEvent(BaseEvent):
    def __init__(self):
        super().__init__(key=UNREACHED_EVENT,
                         value=UNREACHED_EVENT,
                         event_type=UNREACHED_EVENT,
                         namespace=SCHEDULER_NAMESPACE)


class SchedulerInnerEventType(Enum):
    STOP_SCHEDULER = 'STOP_SCHEDULER'
    TASK_STATUS_CHANGED = 'TASK_STATUS_CHANGED'
    TASK_SCHEDULING = 'TASK_SCHEDULING'
    DAG_EXECUTABLE = 'DAG_EXECUTABLE'
    DAG_RUN_FINISHED = 'DAG_RUN_FINISHED'
    EVENT_HANDLE = 'EVENT_HANDLE'
    REQUEST = 'REQUEST'
    RESPONSE = 'RESPONSE'
    STOP_DAG = 'STOP_DAG'
    PARSE_DAG_REQUEST = 'PARSE_DAG_REQUEST'
    PARSE_DAG_RESPONSE = 'PARSE_DAG_RESPONSE'
    PERIODIC_TASK_EVENT = 'PERIODIC_TASK_EVENT'


class SchedulerInnerEvent(object):

    @classmethod
    def to_base_event(cls, event: 'SchedulerInnerEvent') -> BaseEvent:
        raise NotImplementedError()

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'SchedulerInnerEvent':
        raise NotImplementedError()

    def to_event(self)->BaseEvent:
        return self.to_base_event(self)


class PeriodicEvent(SchedulerInnerEvent):

    def __init__(self, run_id, task_id):
        self.run_id = run_id
        self.task_id = task_id

    @classmethod
    def to_base_event(cls, event: 'PeriodicEvent') -> BaseEvent:
        return BaseEvent(key=event.run_id,
                         value=event.task_id,
                         event_type=SchedulerInnerEventType.PERIODIC_TASK_EVENT.value,
                         namespace=SCHEDULER_NAMESPACE)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'PeriodicEvent':
        return PeriodicEvent(event.key, event.value)


class ParseDagRequestEvent(SchedulerInnerEvent):
    def __init__(self, request_id):
        self.request_id = request_id

    @classmethod
    def to_base_event(cls, event: 'ParseDagRequestEvent') -> BaseEvent:
        return BaseEvent(key=str(event.request_id),
                         value='',
                         event_type=SchedulerInnerEventType.PARSE_DAG_REQUEST.value)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'SchedulerInnerEvent':
        return ParseDagRequestEvent(event.key)


class ParseDagResponseEvent(SchedulerInnerEvent):
    def __init__(self, request_id):
        self.request_id = request_id

    @classmethod
    def to_base_event(cls, event: 'ParseDagResponseEvent') -> BaseEvent:
        return BaseEvent(key=str(event.request_id),
                         value='',
                         event_type=SchedulerInnerEventType.PARSE_DAG_RESPONSE.value)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'SchedulerInnerEvent':
        return ParseDagResponseEvent(event.key)


class UserDefineMessageType(Enum):
    RUN_DAG = 'RUN_DAG'
    STOP_DAG_RUN = 'STOP_DAG_RUN'
    EXECUTE_TASK = 'EXECUTE_TASK'


class BaseUserDefineMessage(object):
    def __init__(self, message_type: UserDefineMessageType = None):
        self.message_type = message_type

    def to_json(self) -> str:
        o = {}
        for k, v in self.__dict__.items():
            if k == 'message_type':
                o[k] = v.value
            else:
                o[k] = v
        return json.dumps(o)

    def from_json(self, json_str):
        o = json.loads(json_str)
        for k, v in o.items():
            if k == 'message_type':
                self.__dict__[k] = UserDefineMessageType(v)
            else:
                self.__dict__[k] = v


class RunDagMessage(BaseUserDefineMessage):
    def __init__(self, dag_id):
        super().__init__(UserDefineMessageType.RUN_DAG)
        self.dag_id = dag_id


class StopDagRunMessage(BaseUserDefineMessage):
    def __init__(self, dag_id, dagrun_id):
        super().__init__(UserDefineMessageType.STOP_DAG_RUN)
        self.dag_id = dag_id
        self.dagrun_id = dagrun_id


class ExecuteTaskMessage(BaseUserDefineMessage):
    def __init__(self, dag_id, dagrun_id, task_id, action):
        super().__init__(UserDefineMessageType.EXECUTE_TASK)
        self.dag_id = dag_id
        self.dagrun_id = dagrun_id
        self.task_id = task_id
        self.action = action


class RequestEvent(SchedulerInnerEvent):
    def __init__(self, request_id, body):
        self.request_id = request_id
        self.body = body

    @classmethod
    def to_base_event(cls, event: 'RequestEvent') -> BaseEvent:
        return BaseEvent(key=str(event.request_id), value=event.body, event_type=SchedulerInnerEventType.REQUEST.value)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'SchedulerInnerEvent':
        return RequestEvent(event.key, event.value)


class ResponseEvent(SchedulerInnerEvent):
    def __init__(self, request_id, body):
        self.request_id = request_id
        self.body = body

    @classmethod
    def to_base_event(cls, event: 'ResponseEvent') -> BaseEvent:
        return BaseEvent(key=str(event.request_id), value=event.body, event_type=SchedulerInnerEventType.RESPONSE.value)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'SchedulerInnerEvent':
        return ResponseEvent(event.key, event.value)


class StopSchedulerEvent(SchedulerInnerEvent):
    def __init__(self, job_id):
        super().__init__()
        self.job_id = job_id

    @classmethod
    def to_base_event(cls, event: 'StopSchedulerEvent') -> BaseEvent:
        return BaseEvent(key=str(event.job_id), value='', event_type=SchedulerInnerEventType.STOP_SCHEDULER.value)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'SchedulerInnerEvent':
        return StopSchedulerEvent(int(event.key))


class StopDagEvent(SchedulerInnerEvent):
    def __init__(self, dag_id):
        super().__init__()
        self.dag_id = dag_id

    @classmethod
    def to_base_event(cls, event: 'StopDagEvent') -> BaseEvent:
        return BaseEvent(key=str(event.dag_id),
                         value='',
                         event_type=SchedulerInnerEventType.STOP_DAG.value)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'SchedulerInnerEvent':
        return StopDagEvent(str(event.key))


class TaskStateChangedEvent(SchedulerInnerEvent):

    def __init__(self, task_id, dag_id, execution_date, state, create_time=None):
        super().__init__()
        self.task_id = task_id
        self.dag_id = dag_id
        self.execution_date = execution_date
        self.state = state
        self.create_time = create_time if create_time is not None else int(time.time() * 1000)

    @classmethod
    def to_base_event(cls, event: 'TaskStateChangedEvent') -> BaseEvent:
        o = {}
        for k, v in event.__dict__.items():
            if 'execution_date' == k:
                o[k] = v.strftime(EXECUTION_DATE_FORMAT)
            else:
                o[k] = v
        split_dag_id = event.dag_id.split('.', 1)
        if len(split_dag_id) < 2:
            project_name,  workflow_name = split_dag_id[0], split_dag_id[0]
        else:
            project_name, workflow_name = split_dag_id[0], split_dag_id[1]
        return BaseEvent(key='.'.join([workflow_name, event.task_id]),
                         value=event.state,
                         context=json.dumps(o),
                         event_type=SchedulerInnerEventType.TASK_STATUS_CHANGED.value,
                         namespace=project_name,
                         sender=event.task_id,
                         create_time=event.create_time)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'TaskStateChangedEvent':
        o = json.loads(event.context)
        return TaskStateChangedEvent(task_id=o['task_id'],
                                     dag_id=o['dag_id'],
                                     execution_date=dates.parse_execution_date(o['execution_date']),
                                     state=o['state'],
                                     create_time=event.create_time)


class DagExecutableEvent(SchedulerInnerEvent):
    def __init__(self, dag_id):
        super().__init__()
        self.dag_id = dag_id

    @classmethod
    def to_base_event(cls, event: 'DagExecutableEvent') -> BaseEvent:
        return BaseEvent(key=event.dag_id, value='', event_type=SchedulerInnerEventType.DAG_EXECUTABLE.value,
                         namespace=event.dag_id)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'DagExecutableEvent':
        return DagExecutableEvent(dag_id=event.key)


class DagRunFinishedEvent(SchedulerInnerEvent):
    def __init__(self, run_id):
        super().__init__()
        self.run_id = run_id

    @classmethod
    def to_base_event(cls, event: 'DagRunFinishedEvent') -> BaseEvent:
        return BaseEvent(key=event.run_id, value='', event_type=SchedulerInnerEventType.DAG_RUN_FINISHED.value,
                         namespace=SCHEDULER_NAMESPACE)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'DagRunFinishedEvent':
        return DagRunFinishedEvent(run_id=event.key)


class TaskSchedulingEvent(SchedulerInnerEvent):
    def __init__(self, task_id, dag_id, execution_date, try_number, action: SchedulingAction):
        super().__init__()
        self.task_id = task_id
        self.dag_id = dag_id
        self.execution_date = execution_date
        self.try_number = try_number
        self.action = action

    @classmethod
    def to_base_event(cls, event: 'TaskSchedulingEvent') -> BaseEvent:
        o = {}
        for k, v in event.__dict__.items():
            if 'execution_date' == k:
                o[k] = v.strftime(EXECUTION_DATE_FORMAT)
            elif 'action' == k:
                o[k] = v.value
            else:
                o[k] = v
        return BaseEvent(key=event.dag_id,
                         value=json.dumps(o),
                         event_type=SchedulerInnerEventType.TASK_SCHEDULING.value,
                         namespace=event.dag_id)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'TaskSchedulingEvent':
        o = json.loads(event.value)
        return TaskSchedulingEvent(task_id=o['task_id'],
                                   dag_id=o['dag_id'],
                                   execution_date=dates.parse_execution_date(o['execution_date']),
                                   try_number=int(o['try_number']),
                                   action=SchedulingAction(o['action']))


class EventHandleEvent(SchedulerInnerEvent):
    def __init__(self, dag_id, dag_run_id, task_id, action: SchedulingAction):
        super().__init__()
        self.task_id = task_id
        self.dag_id = dag_id
        self.dag_run_id = dag_run_id
        self.action = action

    @classmethod
    def to_base_event(cls, event: 'EventHandleEvent') -> BaseEvent:
        o = {}
        for k, v in event.__dict__.items():
            if 'action' == k:
                o[k] = v.value
            else:
                o[k] = v
        return BaseEvent(key=event.dag_id,
                         value=json.dumps(o),
                         event_type=SchedulerInnerEventType.EVENT_HANDLE.value,
                         namespace=event.dag_id)

    @classmethod
    def from_base_event(cls, event: BaseEvent) -> 'EventHandleEvent':
        o = json.loads(event.value)
        return EventHandleEvent(task_id=o['task_id'],
                                dag_run_id=o['dag_run_id'],
                                dag_id=o['dag_id'],
                                action=SchedulingAction(o['action']))


class SchedulerInnerEventUtil(object):
    @staticmethod
    def is_inner_event(event: BaseEvent) -> bool:
        try:
            SchedulerInnerEventType(event.event_type)
            return True
        except ValueError as e:
            return False

    @staticmethod
    def event_type(event: BaseEvent) -> SchedulerInnerEventType:
        try:
            return SchedulerInnerEventType(event.event_type)
        except ValueError as e:
            return None

    @staticmethod
    def to_inner_event(event: BaseEvent)->SchedulerInnerEvent:
        event_type = SchedulerInnerEventUtil.event_type(event)
        if SchedulerInnerEventType.STOP_SCHEDULER == event_type:
            return StopSchedulerEvent.from_base_event(event)
        elif SchedulerInnerEventType.TASK_SCHEDULING == event_type:
            return TaskSchedulingEvent.from_base_event(event)
        elif SchedulerInnerEventType.TASK_STATUS_CHANGED == event_type:
            return TaskStateChangedEvent.from_base_event(event)
        elif SchedulerInnerEventType.DAG_EXECUTABLE == event_type:
            return DagExecutableEvent.from_base_event(event)
        elif SchedulerInnerEventType.EVENT_HANDLE == event_type:
            return EventHandleEvent.from_base_event(event)
        elif SchedulerInnerEventType.REQUEST == event_type:
            return RequestEvent.from_base_event(event)
        elif SchedulerInnerEventType.RESPONSE == event_type:
            return ResponseEvent.from_base_event(event)
        elif SchedulerInnerEventType.STOP_DAG == event_type:
            return StopDagEvent.from_base_event(event)
        elif SchedulerInnerEventType.PARSE_DAG_REQUEST == event_type:
            return ParseDagRequestEvent.from_base_event(event)
        elif SchedulerInnerEventType.PARSE_DAG_RESPONSE == event_type:
            return ParseDagResponseEvent.from_base_event(event)
        elif SchedulerInnerEventType.DAG_RUN_FINISHED == event_type:
            return DagRunFinishedEvent.from_base_event(event)
        elif SchedulerInnerEventType.PERIODIC_TASK_EVENT == event_type:
            return PeriodicEvent.from_base_event(event)
        else:
            return None



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Authentication backend that denies all requests"""
from functools import wraps
from typing import Callable, Optional, Tuple, TypeVar, Union, cast

from flask import Response
from requests.auth import AuthBase

CLIENT_AUTH: Optional[Union[Tuple[str, str], AuthBase]] = None


def init_app(_):
    """Initializes authentication"""


T = TypeVar("T", bound=Callable)  # pylint: disable=invalid-name


def requires_authentication(function: T):
    """Decorator for functions that require authentication"""

    @wraps(function)
    def decorated(*args, **kwargs):  # pylint: disable=unused-argument
        return Response("Forbidden", 403)

    return cast(T, decorated)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import List, Optional, Union

from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class GenericTransfer(BaseOperator):
    """
    Moves data from a connection to another, assuming that they both
    provide the required methods in their respective hooks. The source hook
    needs to expose a `get_records` method, and the destination a
    `insert_rows` method.

    This is meant to be used on small-ish datasets that fit in memory.

    :param sql: SQL query to execute against the source database. (templated)
    :type sql: str
    :param destination_table: target table. (templated)
    :type destination_table: str
    :param source_conn_id: source connection
    :type source_conn_id: str
    :param destination_conn_id: source connection
    :type destination_conn_id: str
    :param preoperator: sql statement or list of statements to be
        executed prior to loading the data. (templated)
    :type preoperator: str or list[str]
    """

    template_fields = ('sql', 'destination_table', 'preoperator')
    template_ext = (
        '.sql',
        '.hql',
    )
    ui_color = '#b0f07c'

    @apply_defaults
    def __init__(
        self,
        *,
        sql: str,
        destination_table: str,
        source_conn_id: str,
        destination_conn_id: str,
        preoperator: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.sql = sql
        self.destination_table = destination_table
        self.source_conn_id = source_conn_id
        self.destination_conn_id = destination_conn_id
        self.preoperator = preoperator

    def execute(self, context):
        source_hook = BaseHook.get_hook(self.source_conn_id)

        self.log.info("Extracting data from %s", self.source_conn_id)
        self.log.info("Executing: \n %s", self.sql)
        results = source_hook.get_records(self.sql)

        destination_hook = BaseHook.get_hook(self.destination_conn_id)
        if self.preoperator:
            self.log.info("Running preoperator")
            self.log.info(self.preoperator)
            destination_hook.run(self.preoperator)

        self.log.info("Inserting rows into %s", self.destination_conn_id)
        destination_hook.insert_rows(table=self.destination_table, rows=results)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from flask_appbuilder import AppBuilder

from airflow import settings
from airflow.configuration import conf


def init_appbuilder(app):
    """Init `Flask App Builder <https://flask-appbuilder.readthedocs.io/en/latest/>`__."""
    from airflow.www.security import AirflowSecurityManager

    security_manager_class = app.config.get('SECURITY_MANAGER_CLASS') or AirflowSecurityManager

    if not issubclass(security_manager_class, AirflowSecurityManager):
        raise Exception(
            """Your CUSTOM_SECURITY_MANAGER must now extend AirflowSecurityManager,
             not FAB's security manager."""
        )

    class AirflowAppBuilder(AppBuilder):
        """Custom class to prevent side effects of the session."""

        def _check_and_init(self, baseview):
            if hasattr(baseview, 'datamodel'):
                # Delete sessions if initiated previously to limit side effects. We want to use
                # the current session in the current application.
                baseview.datamodel.session = None
            return super()._check_and_init(baseview)

    AirflowAppBuilder(
        app=app,
        session=settings.Session,
        security_manager_class=security_manager_class,
        base_template='airflow/main.html',
        update_perms=conf.getboolean('webserver', 'UPDATE_FAB_PERMS'),
    )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from airflow.models.taskexecution import TaskExecution

from airflow.jobs.base_job import BaseJob
from airflow.models import (
    Connection,
    DagModel,
    DagRun,
    DagTag,
    Log,
    Pool,
    RenderedTaskInstanceFields,
    SlaMiss,
    TaskFail,
    TaskInstance,
    TaskReschedule,
    Variable,
    XCom,
    errors,
    Message,
)
from airflow.models.dagcode import DagCode
from airflow.models.serialized_dag import SerializedDagModel
from airflow.models.taskstate import TaskState
from airflow.utils.db import add_default_pool_if_not_exists, create_default_connections
from airflow.utils.session import create_session


def clear_db_runs():
    with create_session() as session:
        session.query(DagRun).delete()
        session.query(TaskInstance).delete()


def clear_db_dags():
    with create_session() as session:
        session.query(DagTag).delete()
        session.query(DagModel).delete()
        session.query(DagCode).delete()


def clear_db_serialized_dags():
    with create_session() as session:
        session.query(SerializedDagModel).delete()


def clear_db_sla_miss():
    with create_session() as session:
        session.query(SlaMiss).delete()


def clear_db_errors():
    with create_session() as session:
        session.query(errors.ImportError).delete()


def clear_db_pools():
    with create_session() as session:
        session.query(Pool).delete()
        add_default_pool_if_not_exists(session)


def clear_db_connections(add_default_connections_back=True):
    with create_session() as session:
        session.query(Connection).delete()
        if add_default_connections_back:
            create_default_connections(session)


def clear_db_variables():
    with create_session() as session:
        session.query(Variable).delete()


def clear_db_dag_code():
    with create_session() as session:
        session.query(DagCode).delete()


def set_default_pool_slots(slots):
    with create_session() as session:
        default_pool = Pool.get_default_pool(session)
        default_pool.slots = slots


def clear_rendered_ti_fields():
    with create_session() as session:
        session.query(RenderedTaskInstanceFields).delete()


def clear_db_import_errors():
    with create_session() as session:
        session.query(errors.ImportError).delete()


def clear_db_xcom():
    with create_session() as session:
        session.query(XCom).delete()


def clear_db_logs():
    with create_session() as session:
        session.query(Log).delete()


def clear_db_jobs():
    with create_session() as session:
        session.query(BaseJob).delete()


def clear_db_task_fail():
    with create_session() as session:
        session.query(TaskFail).delete()


def clear_db_task_reschedule():
    with create_session() as session:
        session.query(TaskReschedule).delete()


def clear_db_task_state():
    with create_session() as session:
        session.query(TaskState).delete()


def clear_db_task_execution():
    with create_session() as session:
        session.query(TaskExecution).delete()


def clear_db_message():
    with create_session() as session:
        session.query(Message).delete()


def clear_all():
    clear_db_jobs()
    clear_db_dags()
    clear_db_serialized_dags()
    clear_db_runs()
    clear_db_task_execution()
    clear_db_message()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import List, NamedTuple

from marshmallow import Schema, fields
from marshmallow_sqlalchemy import SQLAlchemySchema, auto_field

from airflow.models.connection import Connection


class ConnectionCollectionItemSchema(SQLAlchemySchema):
    """Schema for a connection item"""

    class Meta:
        """Meta"""

        model = Connection

    connection_id = auto_field('conn_id', required=True)
    conn_type = auto_field(required=True)
    host = auto_field()
    login = auto_field()
    schema = auto_field()
    port = auto_field()


class ConnectionSchema(ConnectionCollectionItemSchema):  # pylint: disable=too-many-ancestors
    """Connection schema"""

    password = auto_field(load_only=True)
    extra = auto_field()


class ConnectionCollection(NamedTuple):
    """List of Connections with meta"""

    connections: List[Connection]
    total_entries: int


class ConnectionCollectionSchema(Schema):
    """Connection Collection Schema"""

    connections = fields.List(fields.Nested(ConnectionCollectionItemSchema))
    total_entries = fields.Int()


connection_schema = ConnectionSchema()
connection_collection_item_schema = ConnectionCollectionItemSchema()
connection_collection_schema = ConnectionCollectionSchema()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
from datetime import datetime as dt

import pendulum
from flask_appbuilder.fieldwidgets import (
    BS3PasswordFieldWidget,
    BS3TextAreaFieldWidget,
    BS3TextFieldWidget,
    Select2Widget,
)
from flask_appbuilder.forms import DynamicForm
from flask_babel import lazy_gettext
from flask_wtf import FlaskForm
from wtforms import widgets
from wtforms.fields import (
    BooleanField,
    Field,
    IntegerField,
    PasswordField,
    SelectField,
    StringField,
    TextAreaField,
)
from wtforms.validators import InputRequired, Optional

from airflow.configuration import conf
from airflow.utils import timezone
from airflow.utils.types import DagRunType
from airflow.www.validators import ValidJson
from airflow.www.widgets import (
    AirflowDateTimePickerROWidget,
    AirflowDateTimePickerWidget,
    BS3TextAreaROWidget,
    BS3TextFieldROWidget,
)


class DateTimeWithTimezoneField(Field):
    """A text field which stores a `datetime.datetime` matching a format."""

    widget = widgets.TextInput()

    def __init__(self, label=None, validators=None, datetime_format='%Y-%m-%d %H:%M:%S%Z', **kwargs):
        super().__init__(label, validators, **kwargs)
        self.format = datetime_format
        self.data = None

    def _value(self):
        if self.raw_data:
            return ' '.join(self.raw_data)
        if self.data:
            return self.data.strftime(self.format)
        return ''

    def process_formdata(self, valuelist):
        if not valuelist:
            return
        date_str = ' '.join(valuelist)
        try:
            # Check if the datetime string is in the format without timezone, if so convert it to the
            # default timezone
            if len(date_str) == 19:
                parsed_datetime = dt.strptime(date_str, '%Y-%m-%d %H:%M:%S')
                default_timezone = self._get_default_timezone()
                self.data = default_timezone.convert(parsed_datetime)
            else:
                self.data = pendulum.parse(date_str)
        except ValueError:
            self.data = None
            raise ValueError(self.gettext('Not a valid datetime value'))

    def _get_default_timezone(self):
        current_timezone = conf.get("core", "default_timezone")
        if current_timezone == "system":
            default_timezone = pendulum.local_timezone()
        else:
            default_timezone = pendulum.timezone(current_timezone)
        return default_timezone


class DateTimeForm(FlaskForm):
    """Date filter form needed for task views"""

    execution_date = DateTimeWithTimezoneField("Execution date", widget=AirflowDateTimePickerWidget())


class DateTimeWithNumRunsForm(FlaskForm):
    """
    Date time and number of runs form for tree view, task duration
    and landing times
    """

    base_date = DateTimeWithTimezoneField(
        "Anchor date", widget=AirflowDateTimePickerWidget(), default=timezone.utcnow()
    )
    num_runs = SelectField(
        "Number of runs",
        default=25,
        choices=(
            (5, "5"),
            (25, "25"),
            (50, "50"),
            (100, "100"),
            (365, "365"),
        ),
    )


class DateTimeWithNumRunsWithDagRunsForm(DateTimeWithNumRunsForm):
    """Date time and number of runs and dag runs form for graph and gantt view"""

    execution_date = SelectField("DAG run")


class DagRunForm(DynamicForm):
    """Form for adding DAG Run"""

    dag_id = StringField(lazy_gettext('Dag Id'), validators=[InputRequired()], widget=BS3TextFieldWidget())
    start_date = DateTimeWithTimezoneField(lazy_gettext('Start Date'), widget=AirflowDateTimePickerWidget())
    end_date = DateTimeWithTimezoneField(lazy_gettext('End Date'), widget=AirflowDateTimePickerWidget())
    run_id = StringField(lazy_gettext('Run Id'), validators=[InputRequired()], widget=BS3TextFieldWidget())
    state = SelectField(
        lazy_gettext('State'),
        choices=(
            ('success', 'success'),
            ('running', 'running'),
            ('failed', 'failed'),
        ),
        widget=Select2Widget(),
        validators=[InputRequired()],
    )
    execution_date = DateTimeWithTimezoneField(
        lazy_gettext('Execution Date'),
        widget=AirflowDateTimePickerWidget(),
        validators=[InputRequired()],
    )
    external_trigger = BooleanField(lazy_gettext('External Trigger'))
    conf = TextAreaField(
        lazy_gettext('Conf'), validators=[ValidJson(), Optional()], widget=BS3TextAreaFieldWidget()
    )

    def populate_obj(self, item):
        """Populates the attributes of the passed obj with data from the form’s fields."""
        super().populate_obj(item)  # pylint: disable=no-member
        item.run_type = DagRunType.from_run_id(item.run_id)
        if item.conf:
            item.conf = json.loads(item.conf)


class DagRunEditForm(DagRunForm):
    """Form for editing DAG Run"""

    dag_id = StringField(lazy_gettext('Dag Id'), validators=[InputRequired()], widget=BS3TextFieldROWidget())
    start_date = DateTimeWithTimezoneField(lazy_gettext('Start Date'), widget=AirflowDateTimePickerROWidget())
    end_date = DateTimeWithTimezoneField(lazy_gettext('End Date'), widget=AirflowDateTimePickerROWidget())
    run_id = StringField(lazy_gettext('Run Id'), validators=[InputRequired()], widget=BS3TextFieldROWidget())
    execution_date = DateTimeWithTimezoneField(
        lazy_gettext('Execution Date'),
        widget=AirflowDateTimePickerROWidget(),
        validators=[InputRequired()],
    )
    conf = TextAreaField(
        lazy_gettext('Conf'), validators=[ValidJson(), Optional()], widget=BS3TextAreaROWidget()
    )


class TaskExecutionEditForm(DynamicForm):
    """Form for editing TaskExecution"""

    dag_id = StringField(lazy_gettext('Dag Id'), validators=[InputRequired()], widget=BS3TextFieldROWidget())
    task_id = StringField(
        lazy_gettext('Task Id'), validators=[InputRequired()], widget=BS3TextFieldROWidget()
    )
    start_date = DateTimeWithTimezoneField(lazy_gettext('Start Date'), widget=AirflowDateTimePickerROWidget())
    end_date = DateTimeWithTimezoneField(lazy_gettext('End Date'), widget=AirflowDateTimePickerROWidget())
    state = SelectField(
        lazy_gettext('State'),
        choices=(
            ('success', 'success'),
            ('running', 'running'),
            ('failed', 'failed'),
            ('up_for_retry', 'up_for_retry'),
        ),
        widget=Select2Widget(),
        validators=[InputRequired()],
    )
    execution_date = DateTimeWithTimezoneField(
        lazy_gettext('Execution Date'),
        widget=AirflowDateTimePickerROWidget(),
        validators=[InputRequired()],
    )


class TaskInstanceEditForm(DynamicForm):
    """Form for editing TaskInstance"""

    dag_id = StringField(lazy_gettext('Dag Id'), validators=[InputRequired()], widget=BS3TextFieldROWidget())
    task_id = StringField(
        lazy_gettext('Task Id'), validators=[InputRequired()], widget=BS3TextFieldROWidget()
    )
    start_date = DateTimeWithTimezoneField(lazy_gettext('Start Date'), widget=AirflowDateTimePickerROWidget())
    end_date = DateTimeWithTimezoneField(lazy_gettext('End Date'), widget=AirflowDateTimePickerROWidget())
    state = SelectField(
        lazy_gettext('State'),
        choices=(
            ('success', 'success'),
            ('running', 'running'),
            ('failed', 'failed'),
            ('up_for_retry', 'up_for_retry'),
        ),
        widget=Select2Widget(),
        validators=[InputRequired()],
    )
    execution_date = DateTimeWithTimezoneField(
        lazy_gettext('Execution Date'),
        widget=AirflowDateTimePickerROWidget(),
        validators=[InputRequired()],
    )


class ConnectionForm(DynamicForm):
    """Form for editing and adding Connection"""

    conn_id = StringField(lazy_gettext('Conn Id'), validators=[InputRequired()], widget=BS3TextFieldWidget())
    description = StringField(lazy_gettext('Description'), widget=BS3TextAreaFieldWidget())
    host = StringField(lazy_gettext('Host'), widget=BS3TextFieldWidget())
    schema = StringField(lazy_gettext('Schema'), widget=BS3TextFieldWidget())
    login = StringField(lazy_gettext('Login'), widget=BS3TextFieldWidget())
    password = PasswordField(lazy_gettext('Password'), widget=BS3PasswordFieldWidget())
    port = IntegerField(lazy_gettext('Port'), validators=[Optional()], widget=BS3TextFieldWidget())
    extra = TextAreaField(lazy_gettext('Extra'), widget=BS3TextAreaFieldWidget())



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Provides lineage support functions"""
import json
import logging
from functools import wraps
from typing import Any, Callable, Dict, Optional, TypeVar, cast

import attr
import jinja2
from cattr import structure, unstructure

from airflow.utils.module_loading import import_string

ENV = jinja2.Environment()

PIPELINE_OUTLETS = "pipeline_outlets"
PIPELINE_INLETS = "pipeline_inlets"
AUTO = "auto"

log = logging.getLogger(__name__)


@attr.s(auto_attribs=True)
class Metadata:
    """Class for serialized entities."""

    type_name: str = attr.ib()
    source: str = attr.ib()
    data: Dict = attr.ib()


def _get_instance(meta: Metadata):
    """Instantiate an object from Metadata"""
    cls = import_string(meta.type_name)
    return structure(meta.data, cls)


def _render_object(obj: Any, context) -> Any:
    """Renders a attr annotated object. Will set non serializable attributes to none"""
    return structure(
        json.loads(
            ENV.from_string(json.dumps(unstructure(obj), default=lambda o: None))
            .render(**context)
            .encode('utf-8')
        ),
        type(obj),
    )


def _to_dataset(obj: Any, source: str) -> Optional[Metadata]:
    """Create Metadata from attr annotated object"""
    if not attr.has(obj):
        return None

    type_name = obj.__module__ + '.' + obj.__class__.__name__
    data = unstructure(obj)

    return Metadata(type_name, source, data)


T = TypeVar("T", bound=Callable)  # pylint: disable=invalid-name


def apply_lineage(func: T) -> T:
    """
    Saves the lineage to XCom and if configured to do so sends it
    to the backend.
    """

    @wraps(func)
    def wrapper(self, context, *args, **kwargs):
        self.log.debug("Lineage called with inlets: %s, outlets: %s", self.inlets, self.outlets)
        ret_val = func(self, context, *args, **kwargs)

        outlets = [unstructure(_to_dataset(x, f"{self.dag_id}.{self.task_id}")) for x in self.outlets]
        inlets = [unstructure(_to_dataset(x, None)) for x in self.inlets]

        if self.outlets:
            self.xcom_push(
                context, key=PIPELINE_OUTLETS, value=outlets, execution_date=context['ti'].execution_date
            )

        if self.inlets:
            self.xcom_push(
                context, key=PIPELINE_INLETS, value=inlets, execution_date=context['ti'].execution_date
            )

        return ret_val

    return cast(T, wrapper)


def prepare_lineage(func: T) -> T:
    """
    Prepares the lineage inlets and outlets. Inlets can be:

    * "auto" -> picks up any outlets from direct upstream tasks that have outlets defined, as such that
      if A -> B -> C and B does not have outlets but A does, these are provided as inlets.
    * "list of task_ids" -> picks up outlets from the upstream task_ids
    * "list of datasets" -> manually defined list of data

    """
    # pylint: disable=protected-access
    @wraps(func)
    def wrapper(self, context, *args, **kwargs):
        from airflow.models.base import Operator

        self.log.debug("Preparing lineage inlets and outlets")

        if isinstance(self._inlets, (str, Operator)) or attr.has(self._inlets):
            self._inlets = [
                self._inlets,
            ]

        if self._inlets and isinstance(self._inlets, list):
            # get task_ids that are specified as parameter and make sure they are upstream
            task_ids = (
                set(filter(lambda x: isinstance(x, str) and x.lower() != AUTO, self._inlets))
                .union(map(lambda op: op.task_id, filter(lambda op: isinstance(op, Operator), self._inlets)))
                .intersection(self.get_flat_relative_ids(upstream=True))
            )

            # pick up unique direct upstream task_ids if AUTO is specified
            if AUTO.upper() in self._inlets or AUTO.lower() in self._inlets:
                task_ids = task_ids.union(task_ids.symmetric_difference(self.upstream_task_ids))

            _inlets = self.xcom_pull(context, task_ids=task_ids, dag_id=self.dag_id, key=PIPELINE_OUTLETS)

            # re-instantiate the obtained inlets
            _inlets = [
                _get_instance(structure(item, Metadata)) for sublist in _inlets if sublist for item in sublist
            ]

            self.inlets.extend(_inlets)
            self.inlets.extend(self._inlets)

        elif self._inlets:
            raise AttributeError("inlets is not a list, operator, string or attr annotated object")

        if not isinstance(self._outlets, list):
            self._outlets = [
                self._outlets,
            ]

        self.outlets.extend(self._outlets)

        # render inlets and outlets
        self.inlets = [_render_object(i, context) for i in self.inlets if attr.has(i)]

        self.outlets = [_render_object(i, context) for i in self.outlets if attr.has(i)]

        self.log.debug("inlets: %s, outlets: %s", self.inlets, self.outlets)
        return func(self, context, *args, **kwargs)

    return cast(T, wrapper)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest
import os
import shutil
import threading
from ai_flow.util.path_util import get_file_dir
from ai_flow.plugin_interface.blob_manager_interface import BlobConfig, BlobManagerFactory


class TestLocalBlobManager(unittest.TestCase):

    def test_project_upload_download_local(self):
        project_path = get_file_dir(__file__)
        config = {'blob_manager_class': 'ai_flow_plugins.blob_manager_plugins.local_blob_manager.LocalBlobManager'}
        blob_config = BlobConfig(config)
        blob_manager = BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                              blob_config.blob_manager_config())
        uploaded_path = blob_manager.upload_project('1', project_path)
        self.assertEqual(uploaded_path, project_path)

        downloaded_path = blob_manager.download_project('1', uploaded_path)
        self.assertEqual(project_path, downloaded_path)

    def test_project_upload_download_local_2(self):
        project_path = get_file_dir(__file__)
        config = {
            'blob_manager_class': 'ai_flow_plugins.blob_manager_plugins.local_blob_manager.LocalBlobManager',
            'blob_manager_config': {
                'local_repository': '/tmp',
                'remote_repository': '/tmp'
            }
        }
        blob_config = BlobConfig(config)
        blob_manager = BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                              blob_config.blob_manager_config())

        uploaded_path = blob_manager.upload_project('1', project_path)

        downloaded_path = blob_manager.download_project('1', uploaded_path)
        self.assertEqual('/tmp/workflow_1_project/blob_manager_plugins', downloaded_path)

    def test_project_download_local_same_time(self):
        project_path = get_file_dir(__file__)
        upload_path = '/tmp/upload'
        download_path = '/tmp/download'
        if not os.path.exists(upload_path):
            os.makedirs(upload_path)
        if not os.path.exists(download_path):
            os.makedirs(download_path)
        config = {
            'blob_manager_class': 'ai_flow_plugins.blob_manager_plugins.local_blob_manager.LocalBlobManager',
            'blob_manager_config': {
                'local_repository': download_path,
                'remote_repository': upload_path
            }
        }
        blob_config = BlobConfig(config)
        blob_manager = BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                              blob_config.blob_manager_config())

        uploaded_path = blob_manager.upload_project('1', project_path)
        download_project_path = "/tmp/download/workflow_1_project"
        if os.path.exists(download_project_path):
            shutil.rmtree(download_project_path)

        def download_project():
            blob_manager.download_project('1', uploaded_path)

        t1 = threading.Thread(target=download_project, args=())
        t1.setDaemon(True)
        t1.start()
        t2 = threading.Thread(target=download_project, args=())
        t2.setDaemon(True)
        t2.start()
        t1.join()
        t2.join()
        downloaded_path = blob_manager.download_project('1', uploaded_path)
        self.assertEqual('/tmp/download/workflow_1_project/blob_manager_plugins', downloaded_path)
        if os.path.exists(upload_path):
            shutil.rmtree(upload_path)
        if os.path.exists(download_path):
            shutil.rmtree(download_path)


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


import datetime
import json
import logging
import traceback
from logging.config import DictConfigurator  # type: ignore
from time import sleep

from sqlalchemy import and_, or_, tuple_

from airflow.exceptions import AirflowException, AirflowTaskTimeout
from airflow.models import BaseOperator, SensorInstance, SkipMixin, TaskInstance
from airflow.settings import LOGGING_CLASS_PATH
from airflow.stats import Stats
from airflow.utils import helpers, timezone
from airflow.utils.decorators import apply_defaults
from airflow.utils.email import send_email
from airflow.utils.log.logging_mixin import set_context
from airflow.utils.module_loading import import_string
from airflow.utils.net import get_hostname
from airflow.utils.session import provide_session
from airflow.utils.state import PokeState, State
from airflow.utils.timeout import timeout

config = import_string(LOGGING_CLASS_PATH)
handler_config = config['handlers']['task']
try:
    formatter_config = config['formatters'][handler_config['formatter']]
except Exception as err:  # pylint: disable=broad-except
    formatter_config = None
    print(err)
dictConfigurator = DictConfigurator(config)


class SensorWork:
    """
    This class stores a sensor work with decoded context value. It is only used
    inside of smart sensor. Create a sensor work based on sensor instance record.
    A sensor work object has the following attributes:
    `dag_id`: sensor_instance dag_id.
    `task_id`: sensor_instance task_id.
    `execution_date`: sensor_instance execution_date.
    `try_number`: sensor_instance try_number
    `poke_context`: Decoded poke_context for the sensor task.
    `execution_context`: Decoded execution_context.
    `hashcode`: This is the signature of poking job.
    `operator`: The sensor operator class.
    `op_classpath`: The sensor operator class path
    `encoded_poke_context`: The raw data from sensor_instance poke_context column.
    `log`: The sensor work logger which will mock the corresponding task instance log.

    :param si: The sensor_instance ORM object.
    """

    def __init__(self, si):
        self.dag_id = si.dag_id
        self.task_id = si.task_id
        self.execution_date = si.execution_date
        self.try_number = si.try_number

        self.poke_context = json.loads(si.poke_context) if si.poke_context else {}
        self.execution_context = json.loads(si.execution_context) if si.execution_context else {}
        try:
            self.log = self._get_sensor_logger(si)
        except Exception as e:  # pylint: disable=broad-except
            self.log = None
            print(e)
        self.hashcode = si.hashcode
        self.start_date = si.start_date
        self.operator = si.operator
        self.op_classpath = si.op_classpath
        self.encoded_poke_context = si.poke_context

    def __eq__(self, other):
        if not isinstance(other, SensorWork):
            return NotImplemented

        return (
            self.dag_id == other.dag_id
            and self.task_id == other.task_id
            and self.execution_date == other.execution_date
            and self.try_number == other.try_number
        )

    @staticmethod
    def create_new_task_handler():
        """
        Create task log handler for a sensor work.
        :return: log handler
        """
        handler_config_copy = {k: handler_config[k] for k in handler_config}
        formatter_config_copy = {k: formatter_config[k] for k in formatter_config}
        handler = dictConfigurator.configure_handler(handler_config_copy)
        formatter = dictConfigurator.configure_formatter(formatter_config_copy)
        handler.setFormatter(formatter)
        return handler

    def _get_sensor_logger(self, si):
        """Return logger for a sensor instance object."""
        # The created log_id is used inside of smart sensor as the key to fetch
        # the corresponding in memory log handler.
        si.raw = False  # Otherwise set_context will fail
        log_id = "-".join(
            [si.dag_id, si.task_id, si.execution_date.strftime("%Y_%m_%dT%H_%M_%S_%f"), str(si.try_number)]
        )
        logger = logging.getLogger('airflow.task' + '.' + log_id)

        if len(logger.handlers) == 0:
            handler = self.create_new_task_handler()
            logger.addHandler(handler)
            set_context(logger, si)

            line_break = "-" * 120
            logger.info(line_break)
            logger.info(
                "Processing sensor task %s in smart sensor service on host: %s", self.ti_key, get_hostname()
            )
            logger.info(line_break)
        return logger

    def close_sensor_logger(self):
        """Close log handler for a sensor work."""
        for handler in self.log.handlers:
            try:
                handler.close()
            except Exception as e:  # pylint: disable=broad-except
                print(e)

    @property
    def ti_key(self):
        """Key for the task instance that maps to the sensor work."""
        return self.dag_id, self.task_id, self.execution_date

    @property
    def cache_key(self):
        """Key used to query in smart sensor for cached sensor work."""
        return self.operator, self.encoded_poke_context


class CachedPokeWork:
    """
    Wrapper class for the poke work inside smart sensor. It saves
    the sensor_task used to poke and recent poke result state.
    state: poke state.
    sensor_task: The cached object for executing the poke function.
    last_poke_time: The latest time this cached work being called.
    to_flush: If we should flush the cached work.
    """

    def __init__(self):
        self.state = None
        self.sensor_task = None
        self.last_poke_time = None
        self.to_flush = False

    def set_state(self, state):
        """
        Set state for cached poke work.
        :param state: The sensor_instance state.
        """
        self.state = state
        self.last_poke_time = timezone.utcnow()

    def clear_state(self):
        """Clear state for cached poke work."""
        self.state = None

    def set_to_flush(self):
        """Mark this poke work to be popped from cached dict after current loop."""
        self.to_flush = True

    def is_expired(self):
        """
        The cached task object expires if there is no poke for 20 minutes.
        :return: Boolean
        """
        return self.to_flush or (timezone.utcnow() - self.last_poke_time).total_seconds() > 1200


class SensorExceptionInfo:
    """
    Hold sensor exception information and the type of exception. For possible transient
    infra failure, give the task more chance to retry before fail it.
    """

    def __init__(
        self,
        exception_info,
        is_infra_failure=False,
        infra_failure_retry_window=datetime.timedelta(minutes=130),
    ):
        self._exception_info = exception_info
        self._is_infra_failure = is_infra_failure
        self._infra_failure_retry_window = infra_failure_retry_window

        self._infra_failure_timeout = None
        self.set_infra_failure_timeout()
        self.fail_current_run = self.should_fail_current_run()

    def set_latest_exception(self, exception_info, is_infra_failure=False):
        """
        This function set the latest exception information for sensor exception. If the exception
        implies an infra failure, this function will check the recorded infra failure timeout
        which was set at the first infra failure exception arrives. There is a 6 hours window
        for retry without failing current run.

        :param exception_info: Details of the exception information.
        :param is_infra_failure: If current exception was caused by transient infra failure.
            There is a retry window _infra_failure_retry_window that the smart sensor will
            retry poke function without failing current task run.
        """
        self._exception_info = exception_info
        self._is_infra_failure = is_infra_failure

        self.set_infra_failure_timeout()
        self.fail_current_run = self.should_fail_current_run()

    def set_infra_failure_timeout(self):
        """
        Set the time point when the sensor should be failed if it kept getting infra
        failure.
        :return:
        """
        # Only set the infra_failure_timeout if there is no existing one
        if not self._is_infra_failure:
            self._infra_failure_timeout = None
        elif self._infra_failure_timeout is None:
            self._infra_failure_timeout = timezone.utcnow() + self._infra_failure_retry_window

    def should_fail_current_run(self):
        """
        :return: Should the sensor fail
        :type: boolean
        """
        return not self.is_infra_failure or timezone.utcnow() > self._infra_failure_timeout

    @property
    def exception_info(self):
        """:return: exception msg."""
        return self._exception_info

    @property
    def is_infra_failure(self):
        """

        :return: If the exception is an infra failure
        :type: boolean
        """
        return self._is_infra_failure

    def is_expired(self):
        """
        :return: If current exception need to be kept.
        :type: boolean
        """
        if not self._is_infra_failure:
            return True
        return timezone.utcnow() > self._infra_failure_timeout + datetime.timedelta(minutes=30)


class SmartSensorOperator(BaseOperator, SkipMixin):
    """
    Smart sensor operators are derived from this class.

    Smart Sensor operators keep refresh a dictionary by visiting DB.
    Taking qualified active sensor tasks. Different from sensor operator,
    Smart sensor operators poke for all sensor tasks in the dictionary at
    a time interval. When a criteria is met or fail by time out, it update
    all sensor task state in task_instance table

    :param soft_fail: Set to true to mark the task as SKIPPED on failure
    :type soft_fail: bool
    :param poke_interval: Time in seconds that the job should wait in
        between each tries.
    :type poke_interval: int
    :param smart_sensor_timeout: Time, in seconds before the internal sensor
        job times out if poke_timeout is not defined.
    :type smart_sensor_timeout: float
    :param shard_min: shard code lower bound (inclusive)
    :type shard_min: int
    :param shard_max: shard code upper bound (exclusive)
    :type shard_max: int
    :param poke_timeout: Time, in seconds before the task times out and fails.
    :type poke_timeout: float
    """

    ui_color = '#e6f1f2'

    @apply_defaults
    def __init__(
        self,
        poke_interval=180,
        smart_sensor_timeout=60 * 60 * 24 * 7,
        soft_fail=False,
        shard_min=0,
        shard_max=100000,
        poke_timeout=6.0,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # super(SmartSensorOperator, self).__init__(*args, **kwargs)
        self.poke_interval = poke_interval
        self.soft_fail = soft_fail
        self.timeout = smart_sensor_timeout
        self._validate_input_values()
        self.hostname = ""

        self.sensor_works = []
        self.cached_dedup_works = {}
        self.cached_sensor_exceptions = {}

        self.max_tis_per_query = 50
        self.shard_min = shard_min
        self.shard_max = shard_max
        self.poke_timeout = poke_timeout

    def _validate_input_values(self):
        if not isinstance(self.poke_interval, (int, float)) or self.poke_interval < 0:
            raise AirflowException("The poke_interval must be a non-negative number")
        if not isinstance(self.timeout, (int, float)) or self.timeout < 0:
            raise AirflowException("The timeout must be a non-negative number")

    @provide_session
    def _load_sensor_works(self, session=None):
        """
        Refresh sensor instances need to be handled by this operator. Create smart sensor
        internal object based on the information persisted in the sensor_instance table.

        """
        SI = SensorInstance
        with Stats.timer() as timer:
            query = (
                session.query(SI)
                .filter(SI.state == State.SENSING)
                .filter(SI.shardcode < self.shard_max, SI.shardcode >= self.shard_min)
            )
            tis = query.all()

        self.log.info("Performance query %s tis, time: %.3f", len(tis), timer.duration)

        # Query without checking dagrun state might keep some failed dag_run tasks alive.
        # Join with DagRun table will be very slow based on the number of sensor tasks we
        # need to handle. We query all smart tasks in this operator
        # and expect scheduler correct the states in _change_state_for_tis_without_dagrun()

        sensor_works = []
        for ti in tis:
            try:
                sensor_works.append(SensorWork(ti))
            except Exception as e:  # pylint: disable=broad-except
                self.log.exception("Exception at creating sensor work for ti %s", ti.key)
                self.log.exception(e, exc_info=True)

        self.log.info("%d tasks detected.", len(sensor_works))

        new_sensor_works = [x for x in sensor_works if x not in self.sensor_works]

        self._update_ti_hostname(new_sensor_works)

        self.sensor_works = sensor_works

    @provide_session
    def _update_ti_hostname(self, sensor_works, session=None):
        """
        Update task instance hostname for new sensor works.

        :param sensor_works: Smart sensor internal object for a sensor task.
        :param session: The sqlalchemy session.
        """
        TI = TaskInstance
        ti_keys = [(x.dag_id, x.task_id, x.execution_date) for x in sensor_works]

        def update_ti_hostname_with_count(count, ti_keys):
            # Using or_ instead of in_ here to prevent from full table scan.
            tis = (
                session.query(TI)
                .filter(or_(tuple_(TI.dag_id, TI.task_id, TI.execution_date) == ti_key for ti_key in ti_keys))
                .all()
            )

            for ti in tis:
                ti.hostname = self.hostname
            session.commit()

            return count + len(ti_keys)

        count = helpers.reduce_in_chunks(update_ti_hostname_with_count, ti_keys, 0, self.max_tis_per_query)
        if count:
            self.log.info("Updated hostname on %s tis.", count)

    @provide_session
    def _mark_multi_state(self, operator, poke_hash, encoded_poke_context, state, session=None):
        """
        Mark state for multiple tasks in the task_instance table to a new state if they have
        the same signature as the poke_hash.

        :param operator: The sensor's operator class name.
        :param poke_hash: The hash code generated from sensor's poke context.
        :param encoded_poke_context: The raw encoded poke_context.
        :param state: Set multiple sensor tasks to this state.
        :param session: The sqlalchemy session.
        """

        def mark_state(ti, sensor_instance):
            ti.state = state
            sensor_instance.state = state
            if state in State.finished:
                ti.end_date = end_date
                ti.set_duration()

        SI = SensorInstance
        TI = TaskInstance

        count_marked = 0
        try:
            query_result = (
                session.query(TI, SI)
                .join(
                    TI,
                    and_(
                        TI.dag_id == SI.dag_id,
                        TI.task_id == SI.task_id,
                        TI.execution_date == SI.execution_date,
                    ),
                )
                .filter(SI.state == State.SENSING)
                .filter(SI.hashcode == poke_hash)
                .filter(SI.operator == operator)
                .with_for_update()
                .all()
            )

            end_date = timezone.utcnow()
            for ti, sensor_instance in query_result:
                if sensor_instance.poke_context != encoded_poke_context:
                    continue

                ti.hostname = self.hostname
                if ti.state == State.SENSING:
                    mark_state(ti=ti, sensor_instance=sensor_instance)
                    count_marked += 1
                else:
                    # ti.state != State.SENSING
                    sensor_instance.state = ti.state

            session.commit()

        except Exception as e:  # pylint: disable=broad-except
            self.log.warning("Exception _mark_multi_state in smart sensor for hashcode %s", str(poke_hash))
            self.log.exception(e, exc_info=True)
        self.log.info("Marked %s tasks out of %s to state %s", count_marked, len(query_result), state)

    @provide_session
    def _retry_or_fail_task(self, sensor_work, error, session=None):
        """
        Change single task state for sensor task. For final state, set the end_date.
        Since smart sensor take care all retries in one process. Failed sensor tasks
        logically experienced all retries and the try_number should be set to max_tries.

        :param sensor_work: The sensor_work with exception.
        :type sensor_work: SensorWork
        :param error: The error message for this sensor_work.
        :type error: str.
        :param session: The sqlalchemy session.
        """

        def email_alert(task_instance, error_info):
            try:
                subject, html_content, _ = task_instance.get_email_subject_content(error_info)
                email = sensor_work.execution_context.get('email')

                send_email(email, subject, html_content)
            except Exception as e:  # pylint: disable=broad-except
                sensor_work.log.warning("Exception alerting email.")
                sensor_work.log.exception(e, exc_info=True)

        def handle_failure(sensor_work, ti):
            if sensor_work.execution_context.get('retries') and ti.try_number <= ti.max_tries:
                # retry
                ti.state = State.UP_FOR_RETRY
                if sensor_work.execution_context.get('email_on_retry') and sensor_work.execution_context.get(
                    'email'
                ):
                    sensor_work.log.info("%s sending email alert for retry", sensor_work.ti_key)
                    email_alert(ti, error)
            else:
                ti.state = State.FAILED
                if sensor_work.execution_context.get(
                    'email_on_failure'
                ) and sensor_work.execution_context.get('email'):
                    sensor_work.log.info("%s sending email alert for failure", sensor_work.ti_key)
                    email_alert(ti, error)

        try:
            dag_id, task_id, execution_date = sensor_work.ti_key
            TI = TaskInstance
            SI = SensorInstance
            sensor_instance = (
                session.query(SI)
                .filter(SI.dag_id == dag_id, SI.task_id == task_id, SI.execution_date == execution_date)
                .with_for_update()
                .first()
            )

            if sensor_instance.hashcode != sensor_work.hashcode:
                # Return without setting state
                return

            ti = (
                session.query(TI)
                .filter(TI.dag_id == dag_id, TI.task_id == task_id, TI.execution_date == execution_date)
                .with_for_update()
                .first()
            )

            if ti:
                if ti.state == State.SENSING:
                    ti.hostname = self.hostname
                    handle_failure(sensor_work, ti)

                    sensor_instance.state = State.FAILED
                    ti.end_date = timezone.utcnow()
                    ti.set_duration()
                else:
                    sensor_instance.state = ti.state
                session.merge(sensor_instance)
                session.merge(ti)
                session.commit()

                sensor_work.log.info(
                    "Task %s got an error: %s. Set the state to failed. Exit.", str(sensor_work.ti_key), error
                )
                sensor_work.close_sensor_logger()

        except AirflowException as e:
            sensor_work.log.warning("Exception on failing %s", sensor_work.ti_key)
            sensor_work.log.exception(e, exc_info=True)

    def _check_and_handle_ti_timeout(self, sensor_work):
        """
        Check if a sensor task in smart sensor is timeout. Could be either sensor operator timeout
        or general operator execution_timeout.

        :param sensor_work: SensorWork
        """
        task_timeout = sensor_work.execution_context.get('timeout', self.timeout)
        task_execution_timeout = sensor_work.execution_context.get('execution_timeout')
        if task_execution_timeout:
            task_timeout = min(task_timeout, task_execution_timeout)

        if (timezone.utcnow() - sensor_work.start_date).total_seconds() > task_timeout:
            error = "Sensor Timeout"
            sensor_work.log.exception(error)
            self._retry_or_fail_task(sensor_work, error)

    def _handle_poke_exception(self, sensor_work):
        """
        Fail task if accumulated exceptions exceeds retries.

        :param sensor_work: SensorWork
        """
        sensor_exception = self.cached_sensor_exceptions.get(sensor_work.cache_key)
        error = sensor_exception.exception_info
        sensor_work.log.exception("Handling poke exception: %s", error)

        if sensor_exception.fail_current_run:
            if sensor_exception.is_infra_failure:
                sensor_work.log.exception(
                    "Task %s failed by infra failure in smart sensor.", sensor_work.ti_key
                )
                # There is a risk for sensor object cached in smart sensor keep throwing
                # exception and cause an infra failure. To make sure the sensor tasks after
                # retry will not fall into same object and have endless infra failure,
                # we mark the sensor task after an infra failure so that  it can be popped
                # before next poke loop.
                cache_key = sensor_work.cache_key
                self.cached_dedup_works[cache_key].set_to_flush()
            else:
                sensor_work.log.exception("Task %s failed by exceptions.", sensor_work.ti_key)
            self._retry_or_fail_task(sensor_work, error)
        else:
            sensor_work.log.info("Exception detected, retrying without failing current run.")
            self._check_and_handle_ti_timeout(sensor_work)

    def _process_sensor_work_with_cached_state(self, sensor_work, state):
        if state == PokeState.LANDED:
            sensor_work.log.info("Task %s succeeded", str(sensor_work.ti_key))
            sensor_work.close_sensor_logger()

        if state == PokeState.NOT_LANDED:
            # Handle timeout if connection valid but not landed yet
            self._check_and_handle_ti_timeout(sensor_work)
        elif state == PokeState.POKE_EXCEPTION:
            self._handle_poke_exception(sensor_work)

    def _execute_sensor_work(self, sensor_work):
        ti_key = sensor_work.ti_key
        log = sensor_work.log or self.log
        log.info("Sensing ti: %s", str(ti_key))
        log.info("Poking with arguments: %s", sensor_work.encoded_poke_context)

        cache_key = sensor_work.cache_key
        if cache_key not in self.cached_dedup_works:
            # create an empty cached_work for a new cache_key
            self.cached_dedup_works[cache_key] = CachedPokeWork()

        cached_work = self.cached_dedup_works[cache_key]

        if cached_work.state is not None:
            # Have a valid cached state, don't poke twice in certain time interval
            self._process_sensor_work_with_cached_state(sensor_work, cached_work.state)
            return

        try:
            with timeout(seconds=self.poke_timeout):
                if self.poke(sensor_work):
                    # Got a landed signal, mark all tasks waiting for this partition
                    cached_work.set_state(PokeState.LANDED)

                    self._mark_multi_state(
                        sensor_work.operator,
                        sensor_work.hashcode,
                        sensor_work.encoded_poke_context,
                        State.SUCCESS,
                    )

                    log.info("Task %s succeeded", str(ti_key))
                    sensor_work.close_sensor_logger()
                else:
                    # Not landed yet. Handle possible timeout
                    cached_work.set_state(PokeState.NOT_LANDED)
                    self._check_and_handle_ti_timeout(sensor_work)

                self.cached_sensor_exceptions.pop(cache_key, None)
        except Exception as e:  # pylint: disable=broad-except
            # The retry_infra_failure decorator inside hive_hooks will raise exception with
            # is_infra_failure == True. Long poking timeout here is also considered an infra
            # failure. Other exceptions should fail.
            is_infra_failure = getattr(e, 'is_infra_failure', False) or isinstance(e, AirflowTaskTimeout)
            exception_info = traceback.format_exc()
            cached_work.set_state(PokeState.POKE_EXCEPTION)

            if cache_key in self.cached_sensor_exceptions:
                self.cached_sensor_exceptions[cache_key].set_latest_exception(
                    exception_info, is_infra_failure=is_infra_failure
                )
            else:
                self.cached_sensor_exceptions[cache_key] = SensorExceptionInfo(
                    exception_info, is_infra_failure=is_infra_failure
                )

            self._handle_poke_exception(sensor_work)

    def flush_cached_sensor_poke_results(self):
        """Flush outdated cached sensor states saved in previous loop."""
        for key, cached_work in self.cached_dedup_works.items():
            if cached_work.is_expired():
                self.cached_dedup_works.pop(key, None)
            else:
                cached_work.state = None

        for ti_key, sensor_exception in self.cached_sensor_exceptions.items():
            if sensor_exception.fail_current_run or sensor_exception.is_expired():
                self.cached_sensor_exceptions.pop(ti_key, None)

    def poke(self, sensor_work):
        """
        Function that the sensors defined while deriving this class should
        override.

        """
        cached_work = self.cached_dedup_works[sensor_work.cache_key]
        if not cached_work.sensor_task:
            init_args = dict(list(sensor_work.poke_context.items()) + [('task_id', sensor_work.task_id)])
            operator_class = import_string(sensor_work.op_classpath)
            cached_work.sensor_task = operator_class(**init_args)

        return cached_work.sensor_task.poke(sensor_work.poke_context)

    def _emit_loop_stats(self):
        try:
            count_poke = 0
            count_poke_success = 0
            count_poke_exception = 0
            count_exception_failures = 0
            count_infra_failure = 0
            for cached_work in self.cached_dedup_works.values():
                if cached_work.state is None:
                    continue
                count_poke += 1
                if cached_work.state == PokeState.LANDED:
                    count_poke_success += 1
                elif cached_work.state == PokeState.POKE_EXCEPTION:
                    count_poke_exception += 1
            for cached_exception in self.cached_sensor_exceptions.values():
                if cached_exception.is_infra_failure and cached_exception.fail_current_run:
                    count_infra_failure += 1
                if cached_exception.fail_current_run:
                    count_exception_failures += 1

            Stats.gauge("smart_sensor_operator.poked_tasks", count_poke)
            Stats.gauge("smart_sensor_operator.poked_success", count_poke_success)
            Stats.gauge("smart_sensor_operator.poked_exception", count_poke_exception)
            Stats.gauge("smart_sensor_operator.exception_failures", count_exception_failures)
            Stats.gauge("smart_sensor_operator.infra_failures", count_infra_failure)
        except Exception as e:  # pylint: disable=broad-except
            self.log.exception("Exception at getting loop stats %s")
            self.log.exception(e, exc_info=True)

    def execute(self, context):
        started_at = timezone.utcnow()

        self.hostname = get_hostname()
        while True:
            poke_start_time = timezone.utcnow()

            self.flush_cached_sensor_poke_results()

            self._load_sensor_works()
            self.log.info("Loaded %s sensor_works", len(self.sensor_works))
            Stats.gauge("smart_sensor_operator.loaded_tasks", len(self.sensor_works))

            for sensor_work in self.sensor_works:
                self._execute_sensor_work(sensor_work)

            duration = (timezone.utcnow() - poke_start_time).total_seconds()

            self.log.info("Taking %s to execute %s tasks.", duration, len(self.sensor_works))

            Stats.timing("smart_sensor_operator.loop_duration", duration)
            Stats.gauge("smart_sensor_operator.executed_tasks", len(self.sensor_works))
            self._emit_loop_stats()

            if duration < self.poke_interval:
                sleep(self.poke_interval - duration)
            if (timezone.utcnow() - started_at).total_seconds() > self.timeout:
                self.log.info("Time is out for smart sensor.")
                return

    def on_kill(self):
        pass


if __name__ == '__main__':
    SmartSensorOperator(task_id='test').execute({})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from sqlalchemy import BigInteger
from sqlalchemy import Column
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base

base = declarative_base()


@compiles(BigInteger, 'sqlite')
def bi_c(element, compiler, **kw):
    return "INTEGER"


class Base:
    __table_args__ = {'sqlite_autoincrement': True}

    """
    Column uuid is SQL model primary key, auto increment, which type is BigInteger, but Integer for SQLite.
    """
    uuid = Column(BigInteger, primary_key=True, autoincrement=True)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
AIFLOW_SERVER_BACKEND_STORE_URI = "AIFLOW_SERVER_BACKEND_STORE_URI"
AIFLOW_SQLALCHEMYSTORE_POOL_SIZE = "AIFLOW_SQLALCHEMYSTORE_POOL_SIZE"
AIFLOW_SQLALCHEMYSTORE_MAX_OVERFLOW = "AIFLOW_SQLALCHEMYSTORE_MAX_OVERFLOW"

MONGO_DB_ALIAS_META_SERVICE = 'meta-service-db'



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Text, Tuple, Optional, Union, List

import grpc

from ai_flow.common.properties import Properties
from ai_flow.endpoint.client.base_client import BaseClient
from ai_flow.endpoint.server import int64Value, stringValue
from ai_flow.meta.metric_meta import MetricMeta, MetricSummary, MetricType
from ai_flow.metric.utils import proto_to_metric_meta, proto_to_metric_summary
from ai_flow.protobuf import metric_service_pb2_grpc
from ai_flow.protobuf.message_pb2 import ReturnCode, MetricTypeProto, MetricSummaryProto, MetricMetaProto
from ai_flow.protobuf.metric_service_pb2 import MetricNameRequest, MetricMetaRequest, \
    ListDatasetMetricMetasRequest, ListModelMetricMetasRequest, MetricSummaryRequest, UuidRequest, \
    ListMetricSummariesRequest


class MetricClient(BaseClient):
    def __init__(self, server_uri):
        super(MetricClient, self).__init__(server_uri)
        channel = grpc.insecure_channel(server_uri)
        self.metric_stub = metric_service_pb2_grpc.MetricServiceStub(channel)

    def register_metric_meta(self,
                             metric_name: Text,
                             metric_type: MetricType,
                             project_name: Text,
                             metric_desc: Optional[Text] = None,
                             dataset_name: Optional[Text] = None,
                             model_name: Optional[Text] = None,
                             job_name: Optional[Text] = None,
                             start_time: int = None,
                             end_time: int = None,
                             uri: Optional[Text] = None,
                             tags: Optional[Text] = None,
                             properties: Properties = None
                             ) -> Tuple[int, Text, Optional[MetricMeta]]:
        """
        Register metric metadata in Metric Center.
        
        :param metric_name: Name of registered metric meta. This is expected to be unique in the backend store.
        :param metric_type: Type of registered metric meta.
        :param project_name: Name of the project associated with the registered metric meta.
        :param metric_desc: (Optional) Description of registered metric meta.
        :param dataset_name: (Optional) Name of the dataset associated with the registered metric meta.
        :param model_name: (Optional) Name of the model associated with the registered metric meta.
        :param job_name: (Optional) Name of the job associated with the registered metric meta.
        :param start_time: (Optional) Start time of registered metric meta.
        :param end_time: (Optional) End time of registered metric meta.
        :param uri: (Optional) Uri of registered metric meta.
        :param tags: (Optional) Tags of registered metric meta.
        :param properties: (Optional) Properties of registered metric meta.
        
        :return: A single :py:class:`ai_flow.meta.metric_meta.MetricMeta` object.
        """
        request = MetricMetaRequest(
            metric_meta=MetricMetaProto(
                metric_name=stringValue(metric_name),
                metric_type=MetricTypeProto.Value(metric_type.value),
                metric_desc=stringValue(metric_desc),
                project_name=stringValue(project_name),
                dataset_name=stringValue(dataset_name),
                model_name=stringValue(model_name),
                job_name=stringValue(job_name),
                start_time=int64Value(start_time),
                end_time=int64Value(end_time),
                uri=stringValue(uri),
                tags=stringValue(tags),
                properties=properties)
        )
        response = self.metric_stub.registerMetricMeta(request)
        if 0 == response.return_code:
            metric_meta_proto = response.metric_meta
            metric_meta = proto_to_metric_meta(metric_meta_proto)
            return response.return_code, response.return_msg, metric_meta
        else:
            return response.return_code, response.return_msg, None

    def update_metric_meta(self,
                           metric_name: Text,
                           project_name: Optional[Text] = None,
                           metric_desc: Optional[Text] = None,
                           dataset_name: Optional[Text] = None,
                           model_name: Optional[Text] = None,
                           job_name: Optional[Text] = None,
                           start_time: int = None,
                           end_time: int = None,
                           uri: Optional[Text] = None,
                           tags: Optional[Text] = None,
                           properties: Properties = None
                           ) -> Tuple[int, Text, Optional[MetricMeta]]:
        """
        Update metric metadata in Metric Center.

        :param metric_name: Name of registered metric meta. This is expected to be unique in the backend store.
        :param project_name: (Optional) Name of the project associated with the registered metric meta.
        :param metric_desc: (Optional) Description of registered metric meta.
        :param dataset_name: (Optional) Name of the dataset associated with the registered metric meta.
        :param model_name: (Optional) Name of the model associated with the registered metric meta.
        :param job_name: (Optional) Name of the job associated with the registered metric meta.
        :param start_time: (Optional) Start time of registered metric meta.
        :param end_time: (Optional) End time of registered metric meta.
        :param uri: (Optional) Uri of registered metric meta.
        :param tags: (Optional) Tags of registered metric meta.
        :param properties: (Optional) Properties of registered metric meta.
        
        :return: A single :py:class:`ai_flow.meta.metric_meta.MetricMeta` object.
        """
        request = MetricMetaRequest(
            metric_meta=MetricMetaProto(
                metric_name=stringValue(metric_name),
                metric_desc=stringValue(metric_desc),
                project_name=stringValue(project_name),
                dataset_name=stringValue(dataset_name),
                model_name=stringValue(model_name),
                job_name=stringValue(job_name),
                start_time=int64Value(start_time),
                end_time=int64Value(end_time),
                uri=stringValue(uri),
                tags=stringValue(tags),
                properties=properties)
        )
        response = self.metric_stub.updateMetricMeta(request)
        if 0 == response.return_code:
            metric_meta_proto = response.metric_meta
            metric_meta = proto_to_metric_meta(metric_meta_proto)
            return response.return_code, response.return_msg, metric_meta
        else:
            return response.return_code, response.return_msg, None

    def delete_metric_meta(self, metric_name: Text) -> bool:
        """
        Delete metric metadata by metric name in Metric Center backend.

        :param metric_name: Name of registered metric meta. This is expected to be unique in the backend store.

        :return: True if successfully deleting the given metric metadata, False if not success.
        """
        request = MetricNameRequest(metric_name=metric_name)
        response = self.metric_stub.deleteMetricMeta(request)
        if response.return_code == str(ReturnCode.SUCCESS):
            return True
        else:
            return False

    def get_metric_meta(self, metric_name: Text) -> Tuple[int, Text, Union[None, MetricMeta]]:
        """
        Get metric metadata detail filter by metric name for Metric Center.

        :param metric_name: Name of registered metric meta. This is expected to be unique in the backend store.

       :return: A single :py:class:`ai_flow.meta.metric_meta.MetricMeta` object.
        """
        request = MetricNameRequest(metric_name=metric_name)
        response = self.metric_stub.getMetricMeta(request)

        if 0 == response.return_code:
            metric_meta_proto = response.metric_meta
            metric_meta = proto_to_metric_meta(metric_meta_proto)
            return response.return_code, response.return_msg, metric_meta
        else:
            return response.return_code, response.return_msg, None

    def list_dataset_metric_metas(self, dataset_name: Text, project_name: Optional[Text] = None) -> Tuple[
            int, Text, Union[None, MetricMeta, List[MetricMeta]]]:
        """
        List of dataset metric metadata filter by dataset name and project name for Metric Center.

        :param dataset_name: Name of the dataset associated with the registered metric meta.
        :param project_name: Name of the project associated with the registered metric meta.

        :return: List of :py:class:`ai_flow.meta.metric_meta.MetricMeta` objects.
        """
        request = ListDatasetMetricMetasRequest(dataset_name=dataset_name, project_name=stringValue(project_name))
        response = self.metric_stub.listDatasetMetricMetas(request)

        if 0 == response.return_code:
            repeated_metric_meta_proto = response.metric_metas
            if 1 == len(repeated_metric_meta_proto):
                metric_meta = proto_to_metric_meta(repeated_metric_meta_proto[0])
                return response.return_code, response.return_msg, metric_meta
            else:
                res = []
                for metric_meta_proto in repeated_metric_meta_proto:
                    res.append(proto_to_metric_meta(metric_meta_proto))
                return response.return_code, response.return_msg, res

        else:
            return response.return_code, response.return_msg, None

    def list_model_metric_metas(self, model_name: Text, project_name: Optional[Text] = None) -> Tuple[
            int, Text, Union[None, MetricMeta, List[MetricMeta]]]:
        """
        List of model metric metadata filter by model name and project name for Metric Center.

        :param model_name: Name of the model associated with the registered metric meta.
        :param project_name: Name of the project associated with the registered metric meta.

        :return: List of :py:class:`ai_flow.meta.metric_meta.MetricMeta` objects.
        """
        request = ListModelMetricMetasRequest(model_name=model_name, project_name=stringValue(project_name))
        response = self.metric_stub.listModelMetricMetas(request)
        if 0 == response.return_code:
            repeated_metric_meta_proto = response.metric_metas
            if 1 == len(repeated_metric_meta_proto):
                metric_meta = proto_to_metric_meta(repeated_metric_meta_proto[0])
                return response.return_code, response.return_msg, metric_meta
            else:
                res = []
                for metric_meta_proto in repeated_metric_meta_proto:
                    res.append(proto_to_metric_meta(metric_meta_proto))
                return response.return_code, response.return_msg, res

        else:
            return response.return_code, response.return_msg, None

    def register_metric_summary(self,
                                metric_name: Text,
                                metric_key: Text,
                                metric_value: Text,
                                metric_timestamp: int,
                                model_version: Optional[Text] = None,
                                job_execution_id: Optional[Text] = None
                                ) -> Tuple[int, Text, Optional[MetricSummary]]:
        """
        Register metric summary in Metric Center.

        :param metric_name: Name of registered metric summary.
        :param metric_key: Key of registered metric summary.
        :param metric_value: Value of registered metric summary.
        :param metric_timestamp: Timestamp of registered metric summary.
        :param model_version: (Optional) Version of the model version associated with the registered metric summary.
        :param job_execution_id: (Optional) ID of the job execution associated with the registered metric summary.

        :return: A single :py:class:`ai_flow.meta.metric_meta.MetricSummary` object.
        """
        request = MetricSummaryRequest(
            metric_summary=MetricSummaryProto(
                metric_name=stringValue(metric_name),
                metric_key=stringValue(metric_key),
                metric_value=stringValue(metric_value),
                metric_timestamp=int64Value(metric_timestamp),
                model_version=stringValue(model_version),
                job_execution_id=stringValue(job_execution_id))
        )
        response = self.metric_stub.registerMetricSummary(request)
        if 0 == response.return_code:
            metric_summary_proto = response.metric_summary
            metric_summary = proto_to_metric_summary(metric_summary_proto)
            return response.return_code, response.return_msg, metric_summary
        else:
            return response.return_code, response.return_msg, None

    def update_metric_summary(self,
                              uuid: int,
                              metric_name: Optional[Text] = None,
                              metric_key: Optional[Text] = None,
                              metric_value: Optional[Text] = None,
                              metric_timestamp: int = None,
                              model_version: Optional[Text] = None,
                              job_execution_id: Optional[Text] = None
                              ) -> Tuple[int, Text, Optional[MetricSummary]]:
        """
        Update metric summary in Metric Center.

        :param uuid: UUID of registered metric summary.
        :param metric_name: Name of registered metric summary.
        :param metric_key: Key of registered metric summary.
        :param metric_value: (Optional) Value of registered metric summary.
        :param metric_timestamp: (Optional) Timestamp of registered metric summary.
        :param model_version: (Optional) Version of the model version associated with the registered metric summary.
        :param job_execution_id: (Optional) ID of the job execution associated with the registered metric summary.

        :return: A single :py:class:`ai_flow.meta.metric_meta.MetricSummary` object.
        """
        request = MetricSummaryRequest(
            metric_summary=MetricSummaryProto(
                uuid=uuid,
                metric_name=stringValue(metric_name),
                metric_key=stringValue(metric_key),
                metric_value=stringValue(metric_value),
                metric_timestamp=int64Value(metric_timestamp),
                model_version=stringValue(model_version),
                job_execution_id=stringValue(job_execution_id))
        )
        response = self.metric_stub.updateMetricSummary(request)
        if 0 == response.return_code:
            metric_summary_proto = response.metric_summary
            metric_summary = proto_to_metric_summary(metric_summary_proto)
            return response.return_code, response.return_msg, metric_summary
        else:
            return response.return_code, response.return_msg, None

    def delete_metric_summary(self, uuid: int) -> bool:
        """
        Delete metric summary by metric uuid in Metric Center backend.

        :param uuid: UUID of registered metric summary.

        :return: Whether to delete the given metric summary.
        """
        request = UuidRequest(uuid=uuid)
        response = self.metric_stub.deleteMetricSummary(request)
        if response.return_code == str(ReturnCode.SUCCESS):
            return True
        else:
            return False

    def get_metric_summary(self, uuid: int) -> Tuple[int, Text, Union[None, MetricSummary]]:
        """
        Get metric summary detail filter by summary uuid for Metric Center.

        :param uuid: UUID of registered metric summary.

        :return: A single :py:class:`ai_flow.meta.metric_meta.MetricSummary` object.
        """
        request = UuidRequest(uuid=uuid)
        response = self.metric_stub.getMetricSummary(request)

        if 0 == response.return_code:
            metric_summary_proto = response.metric_summary
            metric_summary = proto_to_metric_summary(metric_summary_proto)
            return response.return_code, response.return_msg, metric_summary
        else:
            return response.return_code, response.return_msg, None

    def list_metric_summaries(self, metric_name: Optional[Text] = None, metric_key: Optional[Text] = None,
                              model_version: Optional[Text] = None, start_time: int = None, end_time=None) -> Tuple[
            int, Text, Union[None, MetricSummary, List[MetricSummary]]]:
        """
        List of metric summaries filter by metric summary fields for Metric Center.

         :param metric_name: (Optional) Name of filtered metric summary.
        :param metric_key: (Optional) Key of filtered metric summary.
        :param model_version: (Optional) Version of the model version associated with the registered metric summary.
        :param start_time: (Optional) Start time for timestamp filtered metric summary.
        :param end_time: (Optional) End time for timestamp filtered metric summary.

        :return: List of :py:class:`ai_flow.meta.metric_meta.MetricSummary` objects.
        """
        request = ListMetricSummariesRequest(metric_name=stringValue(metric_name), metric_key=stringValue(metric_key),
                                             model_version=stringValue(model_version),
                                             start_time=int64Value(start_time), end_time=int64Value(end_time))
        response = self.metric_stub.listMetricSummaries(request)

        if 0 == response.return_code:
            repeated_metric_summary_proto = response.metric_summaries
            if 1 == len(repeated_metric_summary_proto):
                metric_summary = proto_to_metric_summary(repeated_metric_summary_proto[0])
                return response.return_code, response.return_msg, metric_summary
            else:
                res = []
                for metric_summary_proto in repeated_metric_summary_proto:
                    res.append(proto_to_metric_summary(metric_summary_proto))
                return response.return_code, response.return_msg, res

        else:
            return response.return_code, response.return_msg, None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import abstractmethod, ABC
from typing import Dict, Text, List
import copy
import time
import os
from ai_flow.ai_graph.ai_node import ReadDatasetNode, WriteDatasetNode
from ai_flow.context.workflow_config_loader import current_workflow_config
from ai_flow.workflow.workflow import Workflow
from ai_flow.workflow.job import Job
from ai_flow.ai_graph.ai_graph import AIGraph, AISubGraph
from ai_flow.ai_graph.data_edge import DataEdge
from ai_flow.workflow.control_edge import ControlEdge
from ai_flow.context.project_context import ProjectContext


class JobGenerator(ABC):
    """
    JobGenerator: Convert AISubGraph(ai_flow.ai_graph.ai_graph.AISubGraph) to Job(ai_flow.workflow.job.Job)
    """
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def generate(self, sub_graph: AISubGraph, resource_dir: Text = None) -> Job:
        """
        Convert AISubGraph(ai_flow.ai_graph.ai_graph.AISubGraph) to Job(ai_flow.workflow.job.Job)
        :param sub_graph: An executable Graph composed of AINode and data edges with the same job configuration.
        :param resource_dir: Store the executable files generated during the generation process.
        :return: Job(ai_flow.workflow.job.Job)
        """
        pass


class SplitGraph(AIGraph):
    def __init__(self) -> None:
        super().__init__()
        self.nodes: Dict[Text, AISubGraph] = {}
        self.edges: Dict[Text, List[ControlEdge]] = {}

    def add_node(self, node: AISubGraph):
        self.nodes[node.config.job_name] = node


class GraphSplitter(object):

    def __init__(self) -> None:
        super().__init__()

    def split(self, graph: AIGraph) -> SplitGraph:

        split_graph = SplitGraph()

        for n in graph.nodes.values():
            job_name = n.config.job_name
            if job_name in split_graph.nodes:
                sub_graph = split_graph.nodes.get(job_name)
            else:
                sub_graph = AISubGraph(config=n.config)
                split_graph.add_node(sub_graph)
            sub_graph.add_node(n)

        # add data edge to sub graph
        for sub_graph in split_graph.nodes.values():
            for n in sub_graph.nodes.values():
                if n.node_id in graph.edges:
                    for e in graph.edges[n.node_id]:
                        if isinstance(e, DataEdge):
                            sub_graph.add_edge(n.node_id, e)

        for e in graph.edges:
            for ee in graph.edges[e]:
                if isinstance(ee, ControlEdge):
                    split_graph.add_edge(ee.destination, ee)
        return split_graph


class WorkflowConstructor(object):
    class JobGeneratorRegistry(object):
        def __init__(self) -> None:
            super().__init__()
            self.object_dict: Dict[Text, JobGenerator] = {}

        def register(self, key: Text, value: JobGenerator):
            self.object_dict[key] = value

        def get_object(self, key: Text) -> JobGenerator:
            return self.object_dict[key]

    def __init__(self) -> None:
        super().__init__()
        self.job_generator_registry: WorkflowConstructor.JobGeneratorRegistry \
            = WorkflowConstructor.JobGeneratorRegistry()

    def register_job_generator(self, engine, generator: JobGenerator):
        self.job_generator_registry.register(engine, generator)

    def build_workflow(self, split_graph: SplitGraph, project_context: ProjectContext) -> Workflow:
        workflow = Workflow()
        workflow.workflow_config = current_workflow_config()
        workflow.workflow_snapshot_id = '{}.{}.{}'.format(project_context.project_name, workflow.workflow_name,
                                                          round(time.time() * 1000))
        # add ai_nodes to workflow
        for sub in split_graph.nodes.values():
            if sub.config.job_type not in self.job_generator_registry.object_dict:
                raise Exception("job generator not support job_type {}"
                                .format(sub.config.job_type))
            generator: JobGenerator = self.job_generator_registry \
                .get_object(sub.config.job_type)

            # set job resource dir
            job_resource_dir = os.path.join(project_context.get_generated_path(),
                                            workflow.workflow_snapshot_id,
                                            sub.config.job_name)
            if not os.path.exists(job_resource_dir):
                os.makedirs(job_resource_dir)

            job: Job = generator.generate(sub_graph=sub, resource_dir=job_resource_dir)
            job.resource_dir = job_resource_dir

            # set input output dataset
            for node in sub.nodes.values():
                if isinstance(node, ReadDatasetNode):
                    job.input_dataset_list.append(node.dataset())
                elif isinstance(node, WriteDatasetNode):
                    job.output_dataset_list.append(node.dataset())

            workflow.add_job(job)

        def validate_edge(head, tail):
            if head not in workflow.jobs:
                raise Exception('job: {} is not defined in workflow!'.format(head))
            if tail is not None and tail != '' and tail != '*' and tail not in workflow.jobs:
                raise Exception('job: {} is not defined in workflow!'.format(tail))
        # add edges to workflow
        for edges in split_graph.edges.values():
            for e in edges:
                control_edge = copy.deepcopy(e)
                validate_edge(control_edge.destination, control_edge.source)
                workflow.add_edge(control_edge.destination, control_edge)
        return workflow


class Translator(object):
    """
    Translator translates the user-defined program ai graph(ai_flow.ai_graph.ai_graph.AIGraph)
    into ai flow executable workflow(ai_flow.workflow.workflow.Workflow)
    """
    def __init__(self,
                 graph_splitter: GraphSplitter,
                 workflow_constructor: WorkflowConstructor
                 ) -> None:
        super().__init__()
        self.graph_splitter = graph_splitter
        self.workflow_constructor = workflow_constructor

    def translate(self, graph: AIGraph, project_context: ProjectContext) -> Workflow:
        """
        :param graph: The ai graph(ai_flow.ai_graph.ai_graph.AIGraph)
        :param project_context: The ai flow project context(ai_flow.context.project_context.ProjectContext)
        """
        split_graph = self.graph_splitter.split(graph=graph)
        workflow = self.workflow_constructor.build_workflow(split_graph=split_graph,
                                                            project_context=project_context)
        return workflow


__current_translator__ = Translator(graph_splitter=GraphSplitter(),
                                    workflow_constructor=WorkflowConstructor())


def get_translator() -> Translator:
    return __current_translator__


def register_job_generator(job_type, generator: JobGenerator) -> None:
    __current_translator__.workflow_constructor.register_job_generator(job_type, generator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module is deprecated. Please use `airflow.providers.google.cloud.operators.dataproc`."""

import warnings

from airflow.providers.google.cloud.operators.dataproc import (
    DataprocCreateClusterOperator,
    DataprocDeleteClusterOperator,
    DataprocInstantiateInlineWorkflowTemplateOperator,
    DataprocInstantiateWorkflowTemplateOperator,
    DataprocJobBaseOperator,
    DataprocScaleClusterOperator,
    DataprocSubmitHadoopJobOperator,
    DataprocSubmitHiveJobOperator,
    DataprocSubmitPigJobOperator,
    DataprocSubmitPySparkJobOperator,
    DataprocSubmitSparkJobOperator,
    DataprocSubmitSparkSqlJobOperator,
)

warnings.warn(
    "This module is deprecated. Please use `airflow.providers.google.cloud.operators.dataproc`.",
    DeprecationWarning,
    stacklevel=2,
)


class DataprocClusterCreateOperator(DataprocCreateClusterOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocCreateClusterOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dataproc.DataprocCreateClusterOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataprocClusterDeleteOperator(DataprocDeleteClusterOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocDeleteClusterOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dataproc.DataprocDeleteClusterOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataprocClusterScaleOperator(DataprocScaleClusterOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocScaleClusterOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dataproc.DataprocScaleClusterOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataProcHadoopOperator(DataprocSubmitHadoopJobOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHadoopJobOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHadoopJobOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataProcHiveOperator(DataprocSubmitHiveJobOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHiveJobOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHiveJobOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataProcJobBaseOperator(DataprocJobBaseOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocJobBaseOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dataproc.DataprocJobBaseOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataProcPigOperator(DataprocSubmitPigJobOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPigJobOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPigJobOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataProcPySparkOperator(DataprocSubmitPySparkJobOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPySparkJobOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPySparkJobOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataProcSparkOperator(DataprocSubmitSparkJobOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkJobOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkJobOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataProcSparkSqlOperator(DataprocSubmitSparkSqlJobOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkSqlJobOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkSqlJobOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataprocWorkflowTemplateInstantiateInlineOperator(DataprocInstantiateInlineWorkflowTemplateOperator):
    """
    This class is deprecated.
    Please use
    `airflow.providers.google.cloud.operators.dataproc.DataprocInstantiateInlineWorkflowTemplateOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.dataproc
            .DataprocInstantiateInlineWorkflowTemplateOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class DataprocWorkflowTemplateInstantiateOperator(DataprocInstantiateWorkflowTemplateOperator):
    """
    This class is deprecated.
    Please use
    `airflow.providers.google.cloud.operators.dataproc.DataprocInstantiateWorkflowTemplateOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.dataproc
            .DataprocInstantiateWorkflowTemplateOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Callable, Dict, List, Tuple, Union

AIRFLOW_VAR_NAME_FORMAT_MAPPING = {
    'AIRFLOW_CONTEXT_DAG_ID': {'default': 'airflow.ctx.dag_id', 'env_var_format': 'AIRFLOW_CTX_DAG_ID'},
    'AIRFLOW_CONTEXT_TASK_ID': {'default': 'airflow.ctx.task_id', 'env_var_format': 'AIRFLOW_CTX_TASK_ID'},
    'AIRFLOW_CONTEXT_EXECUTION_DATE': {
        'default': 'airflow.ctx.execution_date',
        'env_var_format': 'AIRFLOW_CTX_EXECUTION_DATE',
    },
    'AIRFLOW_CONTEXT_DAG_RUN_ID': {
        'default': 'airflow.ctx.dag_run_id',
        'env_var_format': 'AIRFLOW_CTX_DAG_RUN_ID',
    },
    'AIRFLOW_CONTEXT_DAG_OWNER': {
        'default': 'airflow.ctx.dag_owner',
        'env_var_format': 'AIRFLOW_CTX_DAG_OWNER',
    },
    'AIRFLOW_CONTEXT_DAG_EMAIL': {
        'default': 'airflow.ctx.dag_email',
        'env_var_format': 'AIRFLOW_CTX_DAG_EMAIL',
    },
}


def context_to_airflow_vars(context, in_env_var_format=False):
    """
    Given a context, this function provides a dictionary of values that can be used to
    externally reconstruct relations between dags, dag_runs, tasks and task_instances.
    Default to abc.def.ghi format and can be made to ABC_DEF_GHI format if
    in_env_var_format is set to True.

    :param context: The context for the task_instance of interest.
    :type context: dict
    :param in_env_var_format: If returned vars should be in ABC_DEF_GHI format.
    :type in_env_var_format: bool
    :return: task_instance context as dict.
    """
    params = {}
    if in_env_var_format:
        name_format = 'env_var_format'
    else:
        name_format = 'default'
    task = context.get('task')
    if task and task.email:
        if isinstance(task.email, str):
            params[AIRFLOW_VAR_NAME_FORMAT_MAPPING['AIRFLOW_CONTEXT_DAG_EMAIL'][name_format]] = task.email
        elif isinstance(task.email, list):
            # os env variable value needs to be string
            params[AIRFLOW_VAR_NAME_FORMAT_MAPPING['AIRFLOW_CONTEXT_DAG_EMAIL'][name_format]] = ','.join(
                task.email
            )
    if task and task.owner:
        if isinstance(task.owner, str):
            params[AIRFLOW_VAR_NAME_FORMAT_MAPPING['AIRFLOW_CONTEXT_DAG_OWNER'][name_format]] = task.owner
        elif isinstance(task.owner, list):
            # os env variable value needs to be string
            params[AIRFLOW_VAR_NAME_FORMAT_MAPPING['AIRFLOW_CONTEXT_DAG_OWNER'][name_format]] = ','.join(
                task.owner
            )
    task_instance = context.get('task_instance')
    if task_instance and task_instance.dag_id:
        params[AIRFLOW_VAR_NAME_FORMAT_MAPPING['AIRFLOW_CONTEXT_DAG_ID'][name_format]] = task_instance.dag_id
    if task_instance and task_instance.task_id:
        params[
            AIRFLOW_VAR_NAME_FORMAT_MAPPING['AIRFLOW_CONTEXT_TASK_ID'][name_format]
        ] = task_instance.task_id
    if task_instance and task_instance.execution_date:
        params[
            AIRFLOW_VAR_NAME_FORMAT_MAPPING['AIRFLOW_CONTEXT_EXECUTION_DATE'][name_format]
        ] = task_instance.execution_date.isoformat()
    dag_run = context.get('dag_run')
    if dag_run and dag_run.run_id:
        params[AIRFLOW_VAR_NAME_FORMAT_MAPPING['AIRFLOW_CONTEXT_DAG_RUN_ID'][name_format]] = dag_run.run_id
    return params


def determine_kwargs(func: Callable, args: Union[Tuple, List], kwargs: Dict) -> Dict:
    """
    Inspect the signature of a given callable to determine which arguments in kwargs need
    to be passed to the callable.

    :param func: The callable that you want to invoke
    :param args: The positional arguments that needs to be passed to the callable, so we
        know how many to skip.
    :param kwargs: The keyword arguments that need to be filtered before passing to the callable.
    :return: A dictionary which contains the keyword arguments that are compatible with the callable.
    """
    import inspect
    import itertools

    signature = inspect.signature(func)
    has_kwargs = any(p.kind == p.VAR_KEYWORD for p in signature.parameters.values())

    for name in itertools.islice(signature.parameters.keys(), len(args)):
        # Check if args conflict with names in kwargs
        if name in kwargs:
            raise ValueError(f"The key {name} in args is part of kwargs and therefore reserved.")

    if has_kwargs:
        # If the callable has a **kwargs argument, it's ready to accept all the kwargs.
        return kwargs

    # If the callable has no **kwargs argument, it only wants the arguments it requested.
    return {key: kwargs[key] for key in signature.parameters if key in kwargs}


def make_kwargs_callable(func: Callable) -> Callable:
    """
    Make a new callable that can accept any number of positional or keyword arguments
    but only forwards those required by the given callable func.
    """
    import functools

    @functools.wraps(func)
    def kwargs_func(*args, **kwargs):
        kwargs = determine_kwargs(func, args, kwargs)
        return func(*args, **kwargs)

    return kwargs_func



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest
import os

from ai_flow.util.path_util import get_file_dir

from ai_flow.context.project_context import init_project_context, current_project_config, build_project_context


class TestProjectContext(unittest.TestCase):

    def test_project_init_context(self):
        project_path = os.path.dirname(__file__)
        init_project_context(project_path)
        self.assertEqual('test_project', current_project_config().get_project_name())
        self.assertEqual('a', current_project_config().get('a'))

    def test_build_project_context(self):
        project_path = get_file_dir(__file__)
        project_context = build_project_context(project_path)
        self.assertTrue(project_context.get_project_config_file().endswith('project.yaml'))
        self.assertTrue(project_context.get_generated_path().endswith('generated'))
        self.assertTrue(project_context.get_workflows_path().endswith('workflows'))
        self.assertTrue(project_context.get_dependencies_path().endswith('dependencies'))
        self.assertTrue(project_context.get_python_dependencies_path().endswith('python'))
        self.assertTrue(project_context.get_go_dependencies_path().endswith('go'))
        self.assertTrue(project_context.get_jar_dependencies_path().endswith('jar'))
        self.assertTrue(project_context.get_resources_path().endswith('resources'))


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Add event_based_scheduler tables

Revision ID: d47076ca1426
Revises: 98271e7606e2
Create Date: 2020-12-29 16:00:00.

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
from sqlalchemy.engine.reflection import Inspector

revision = 'd47076ca1426'
down_revision = '98271e7606e2'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()  # pylint: disable=no-member
    is_mysql = bool(conn.dialect.name == "mysql")
    is_sqlite = bool(conn.dialect.name == "sqlite")
    datetime = sa.DATETIME() if not is_mysql else mysql.DATETIME(fsp=6)
    inspector = Inspector.from_engine(conn)
    tables = inspector.get_table_names()


    if is_sqlite:
        op.execute("PRAGMA foreign_keys=off")

    """Apply Add event_relationships Column to dag table"""
    with op.batch_alter_table('serialized_dag', schema=None) as batch_op:
        batch_op.add_column(sa.Column('event_relationships', sa.String(10240), nullable=True))

    """Apply Add scheduling_job_id,event_ack_id Column to dag_run table"""
    with op.batch_alter_table('dag_run', schema=None) as batch_op:
        batch_op.add_column(sa.Column('scheduling_job_id', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('event_ack_id', sa.BigInteger(), nullable=True))

    """Add task_state table"""
    if 'task_state' not in tables:
        op.create_table(
            'task_state',
            sa.Column('task_id', sa.String(length=250), nullable=False),
            sa.Column('dag_id', sa.String(length=250), nullable=False),
            sa.Column('execution_date', datetime, nullable=False),
            sa.Column('task_state', sa.PickleType(), nullable=True),
            sa.PrimaryKeyConstraint('task_id', 'dag_id', 'execution_date')
        )

    """Add task_execution table"""
    if 'task_execution' not in tables:
        op.create_table(
            'task_execution',
            sa.Column('task_id', sa.String(length=250), nullable=False),
            sa.Column('dag_id', sa.String(length=250), nullable=False),
            sa.Column('execution_date', datetime, nullable=False),
            sa.Column('seq_num', sa.Integer(), nullable=False),
            sa.Column('start_date', datetime, nullable=True),
            sa.Column('end_date', datetime, nullable=True),
            sa.Column('duration', sa.Float(), nullable=True),
            sa.Column('state', sa.String(length=20), nullable=True),
            sa.Column('hostname', sa.String(length=1000), nullable=True),
            sa.Column('unixname', sa.String(length=1000), nullable=True),
            sa.Column('job_id', sa.Integer(), nullable=True),
            sa.Column('pool', sa.String(length=50), nullable=True),
            sa.Column('pool_slots', sa.Integer(), default=1),
            sa.Column('queue', sa.String(length=256), nullable=True),
            sa.Column('priority_weight', sa.Integer(), nullable=True),
            sa.Column('operator', sa.String(length=1000), nullable=True),
            sa.Column('queued_dttm', datetime, nullable=True),
            sa.Column('queued_by_job_id', sa.Integer(), nullable=True),
            sa.Column('pid', sa.Integer(), nullable=True),
            sa.Column('executor_config', sa.PickleType(), nullable=True),
            sa.PrimaryKeyConstraint('task_id', 'dag_id', 'execution_date', 'seq_num')
        )

    """Add message table"""
    if 'message' not in tables:
        op.create_table(
            'message',
            sa.Column('id', sa.Integer(), nullable=False),
            sa.Column('message_type', sa.String(length=250), nullable=False),
            sa.Column('data', sa.PickleType(), nullable=True),
            sa.Column('state', sa.String(length=250), nullable=False),
            sa.Column('scheduling_job_id', sa.Integer(), nullable=True),
            sa.Column('queue_time', sa.DateTime(), nullable=True),
            sa.Column('complete_time', sa.DateTime(), nullable=True),
            sa.PrimaryKeyConstraint('id')
        )
        op.create_index('ti_queue_time', 'message', ['queue_time'], unique=False)
        op.create_index('ti_state', 'message', ['state'], unique=False)

def downgrade():
    conn = op.get_bind()  # pylint: disable=no-member
    is_sqlite = bool(conn.dialect.name == "sqlite")

    if is_sqlite:
        op.execute("PRAGMA foreign_keys=off")

    with op.batch_alter_table('serialized_dag', schema=None) as batch_op:
        batch_op.drop_column('event_relationships')

    with op.batch_alter_table('dag_run', schema=None) as batch_op:
        batch_op.drop_column('scheduling_job_id')
        batch_op.drop_column('event_ack_id')

    op.drop_table('task_state')
    op.drop_table('task_execution')
    op.drop_table('message')



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from notification_service.base_notification import BaseEvent
from sqlalchemy import (
    Column, String, ForeignKey, Integer, PrimaryKeyConstraint, BigInteger, UniqueConstraint, Text, Boolean)
from sqlalchemy.orm import relationship, backref
from mongoengine import (Document, StringField, IntField, LongField, ReferenceField,
                         BooleanField, ListField, ObjectIdField, SequenceField)

from ai_flow.meta.metric_meta import MetricType
from ai_flow.model_center.entity.model_version_detail import ModelVersionDetail
from ai_flow.model_center.entity.model_version_stage import STAGE_DELETED, STAGE_GENERATED
from ai_flow.model_center.entity.model_version_status import ModelVersionStatus
from ai_flow.model_center.entity.registered_model_detail import RegisteredModelDetail
from ai_flow.store import MONGO_DB_ALIAS_META_SERVICE
from ai_flow.store.db.base_model import base, Base


class SqlDataset(base, Base):
    """
    SQL table of dataset in metadata backend storage.
    """
    __tablename__ = 'dataset'

    name = Column(String(255), unique=True, nullable=False)
    format = Column(String(256))
    description = Column(String(1000))
    uri = Column(String(1000))
    create_time = Column(BigInteger)
    update_time = Column(BigInteger)
    properties = Column(String(1000))
    name_list = Column(String(1000))
    type_list = Column(String(1000))
    catalog_name = Column(String(1000))
    catalog_type = Column(String(1000))
    catalog_database = Column(String(1000))
    catalog_connection_uri = Column(String(1000))
    catalog_table = Column(String(1000))
    is_deleted = Column(String(256), default='False')

    def __repr__(self):
        return '<Document dataset ({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})>'.format(
            self.uuid,
            self.name,
            self.properties,
            self.name_list,
            self.type_list,
            self.format,
            self.uri,
            self.catalog_name,
            self.catalog_type,
            self.catalog_database,
            self.catalog_table)


class SqlProject(base, Base):
    """
    SQL table of project in metadata backend storage.
    """
    __tablename__ = 'project'

    name = Column(String(255), unique=True)
    properties = Column(String(1000))
    uri = Column(String(1000))
    is_deleted = Column(String(256), default='False')

    def __repr__(self):
        return '<project ({}, {}, {}, {})>'.format(self.uuid, self.name, self.properties, self.uri)


class SqlProjectSnapshot(base, Base):
    """
    SQL table of project snapshot in metadata backend storage
    """
    __tablename__ = 'project_snapshot'

    project_id = Column(BigInteger, ForeignKey('project.uuid', onupdate='cascade'))
    signature = Column(String(255))
    create_time = Column(BigInteger)

    project = relationship("SqlProject", backref=backref('project_snapshot', cascade='all'))


class SqlModelRelation(base, Base):
    """
    SQL table of model relation in metadata backend storage.
    """
    __tablename__ = 'model_relation'

    name = Column(String(255), unique=True)
    project_id = Column(BigInteger, ForeignKey('project.uuid', onupdate='cascade'))
    is_deleted = Column(String(256), default='False')

    project = relationship("SqlProject", backref=backref('model_relation', cascade='all'))

    def __repr__(self):
        return '<model_relation ({}, {}, {})>'.format(self.uuid, self.name, self.project_id)


class SqlWorkflow(base, Base):
    """
    SQL table of workflow in metadata backend storage.
    """
    __tablename__ = 'workflow'

    name = Column(String(255), nullable=False)
    project_id = Column(BigInteger, ForeignKey('project.uuid', onupdate='cascade'))
    properties = Column(String(1000))
    create_time = Column(BigInteger)
    update_time = Column(BigInteger)
    is_deleted = Column(Boolean, default=False)

    UniqueConstraint(project_id, name)

    project = relationship("SqlProject", backref=backref('workflow', cascade='all'))

    def __repr__(self):
        return '<workflow ({}, {}, {}, {}, {}, {}, {})>'.format(self.uuid, self.name, self.project_id, self.properties,
                                                                self.create_time, self.update_time, self.is_deleted)


class SqlModelVersionRelation(base):
    """
    SQL table of model version relation in metadata backend storage.
    """
    __tablename__ = 'model_version_relation'

    version = Column(String(255), primary_key=True)
    model_id = Column(BigInteger, ForeignKey('model_relation.uuid', onupdate='cascade'), primary_key=True)
    project_snapshot_id = Column(BigInteger, ForeignKey('project_snapshot.uuid', onupdate='cascade'))
    is_deleted = Column(String(256), default='False')

    model_relation = relationship("SqlModelRelation", backref=backref('model_version_relation', cascade='all'))
    project_snapshot = relationship("SqlProjectSnapshot", backref=backref('project_snapshot', cascade='all'))

    def __repr__(self):
        return '<model_version_relation ({}, {}, {})>'.format(self.version, self.model_id,
                                                              self.workflow_execution_id)


class SqlArtifact(base, Base):
    """
    SQL table of artifact in metadata backend storage.
    """

    __tablename__ = 'artifact'

    name = Column(String(255), unique=True, nullable=False)
    artifact_type = Column(String(256))
    description = Column(String(1000))
    uri = Column(String(1000))
    create_time = Column(BigInteger)
    update_time = Column(BigInteger)
    properties = Column(String(1000))
    is_deleted = Column(String(256), default='False')

    def __repr__(self):
        return '<SqlArtifact ({}, {}, {}, {}, {}, {}, {}, {}, {})>'.format(self.uuid, self.name, self.artifact_type,
                                                                           self.description, self.uri,
                                                                           self.create_time, self.update_time,
                                                                           self.properties, self.is_deleted)


class SqlRegisteredModel(base):
    """
    SQL model of registered model in Model Center backend storage.
    """
    __tablename__ = 'registered_model'

    model_name = Column(String(255), unique=True, nullable=False)
    model_desc = Column(String(1000), nullable=True)

    __table_args__ = (
        PrimaryKeyConstraint('model_name', name='registered_model_pk'),
    )

    def __repr__(self):
        return '<SqlRegisteredModel ({}, {}, {})>'.format(self.model_name, self.model_desc)

    # entity mappers
    def to_meta_entity(self):
        return RegisteredModelDetail(self.model_name, self.model_desc)

    def to_detail_entity(self):
        # SqlRegisteredModel has backref to all "model_version". Filter latest version of registered model.
        latest_version = None
        for model_version in reversed(self.model_version):
            if model_version.current_stage != STAGE_DELETED:
                latest_version = model_version.to_meta_entity()
                break
        return RegisteredModelDetail(self.model_name, self.model_desc, latest_version)


class SqlModelVersion(base):
    """
    SQL model of model version in Model Center backend storage.
    """
    __tablename__ = 'model_version'

    model_name = Column(String(255), ForeignKey('registered_model.model_name', onupdate='cascade', ondelete='cascade'))
    model_version = Column(String(10), nullable=False)
    model_path = Column(String(500), nullable=True, default=None)
    model_type = Column(String(500), nullable=True, default=None)
    version_desc = Column(String(1000), nullable=True)
    version_status = Column(String(20),
                            default=ModelVersionStatus.to_string(ModelVersionStatus.READY))
    current_stage = Column(String(20), default=STAGE_GENERATED)

    # linked entities
    registered_model = relationship('SqlRegisteredModel', backref=backref('model_version', cascade='all'))

    __table_args__ = (
        PrimaryKeyConstraint('model_name', 'model_version', 'current_stage', name='model_version_pk'),
    )

    def __repr__(self):
        return '<SqlModelVersion ({}, {}, {}, {}, {}, {}, {})>'.format(self.model_name, self.model_version,
                                                                       self.model_path, self.model_type,
                                                                       self.version_desc, self.version_status,
                                                                       self.current_stage)

    # entity mappers
    def to_meta_entity(self):
        return ModelVersionDetail(self.model_name, self.model_version,
                                  self.model_path, self.model_type, self.version_desc,
                                  self.version_status, self.current_stage)


class SqlEvent(base):
    """
    SQL model of event in Notification Service backend storage.
    """
    __tablename__ = 'event'

    __table_args__ = {'sqlite_autoincrement': True}

    id = Column(BigInteger, primary_key=True, autoincrement=True)
    key = Column(String(1024), nullable=False)
    version = Column(Integer, nullable=False)
    value = Column(String(1024))
    event_type = Column(String(256))
    create_time = Column(BigInteger)

    def __repr__(self):
        return '<SqlEvent ({}, {}, {}, {}, {} {})>'.format(self.id, self.key, self.value, self.version,
                                                           self.event_type, self.create_time)

    # entity mappers
    def to_event(self):
        return BaseEvent(self.key, self.value, self.event_type, self.version, self.create_time, self.id)


class SqlMetricMeta(base, Base):
    """
    SQL model of metric meta
    """
    __tablename__ = 'metric_meta'

    metric_name = Column(String(256), unique=True, nullable=False)
    metric_type = Column(String(256), default=MetricType.DATASET.value)
    metric_desc = Column(String(4096))
    project_name = Column(String(256), nullable=False)
    dataset_name = Column(String(256))
    model_name = Column(String(256))
    job_name = Column(String(256))
    start_time = Column(BigInteger)
    end_time = Column(BigInteger)
    uri = Column(String(1024))
    tags = Column(String(256))
    properties = Column(String(1024))
    is_deleted = Column(String(128), default='False')

    def __repr__(self):
        return '<SqlMetricMeta ({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})>' \
            .format(self.uuid, self.metric_name, self.metric_type, self.metric_desc, self.project_name,
                    self.dataset_name, self.model_name, self.start_time, self.end_time, self.uri, self.tags,
                    self.properties)


class SqlMetricSummary(base, Base):
    """
    SQL model of metric summary
    """
    __tablename__ = 'metric_summary'

    metric_name = Column(String(256), ForeignKey('metric_meta.metric_name', onupdate='cascade'), nullable=False)
    metric_key = Column(String(256), nullable=False)
    metric_value = Column(String(2048), nullable=False)
    metric_timestamp = Column(BigInteger, nullable=False)
    model_version = Column(String(256))
    job_execution_id = Column(String(256))
    is_deleted = Column(String(128), default='False')

    metric_summary = relationship("SqlMetricMeta", backref=backref('metric_summary', cascade='all'))

    def __repr__(self):
        return '<SqlMetricSummary ({}, {}, {}, {}, {}, {}, {})>'.format(self.uuid, self.metric_name, self.metric_key,
                                                                        self.metric_value, self.metric_timestamp,
                                                                        self.model_version, self.job_execution_id)


class SqlMember(base):
    """
    SQL model of cluster member.
    """
    __tablename__ = 'aiflow_member'
    id = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)
    version = Column(BigInteger(), nullable=False)
    server_uri = Column(String(767), nullable=False, unique=True)
    update_time = Column(BigInteger(), nullable=False)
    uuid = Column(String(128), nullable=False, unique=True)

    def __repr__(self):
        return '<SqlMember ({}, {}, {}, {}, {})>'.format(
            self.id, self.version, self.server_uri, self.update_time, self.uuid)


class MongoDataset(Document):
    """
    Document of dataset in metadata backend storage.
    """

    uuid = SequenceField(db_alias=MONGO_DB_ALIAS_META_SERVICE)
    name = StringField(max_length=255, required=True, unique=True)
    format = StringField(max_length=256)
    description = StringField(max_length=1000)
    uri = StringField(max_length=1000)
    create_time = LongField()
    update_time = LongField()
    properties = StringField(max_length=1000)
    name_list = StringField(max_length=1000)
    type_list = StringField(max_length=1000)
    catalog_name = StringField(max_length=1000)
    catalog_type = StringField(max_length=1000)
    catalog_database = StringField(max_length=1000)
    catalog_connection_uri = StringField(max_length=1000)
    catalog_table = StringField(max_length=1000)
    is_deleted = BooleanField(default=False)

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document dataset ({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})>'.format(
            self.uuid,
            self.name,
            self.properties,
            self.name_list,
            self.type_list,
            self.format,
            self.uri,
            self.catalog_name,
            self.catalog_type,
            self.catalog_database,
            self.catalog_table)


class MongoModelVersionRelation(Document):
    """
    Document of model version relation in metadata backend storage.
    """

    version = StringField(max_length=255, required=True, unique=True)
    model_id = IntField()
    project_snapshot_id = IntField()
    version_model_id_unique = StringField(max_length=1000, required=True, unique=True)
    is_deleted = BooleanField(default=False)

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __init__(self, *args, **kwargs):
        version = kwargs['version']
        model_id = kwargs['model_id']
        kwargs['version_model_id_unique'] = f'{version}-{model_id}'
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return '<Document ModelVersionRelation ({}, {}, {})>'.format(
            self.version,
            self.model_id,
            self.project_snapshot_id)


class MongoModelRelation(Document):
    """
    Document of model relation in metadata backend storage.
    """

    uuid = SequenceField(db_alias=MONGO_DB_ALIAS_META_SERVICE)
    name = StringField(max_length=255, required=True, unique=True)
    project_id = IntField()
    is_deleted = BooleanField(default=False)

    model_version_relation = ListField(ReferenceField(MongoModelVersionRelation))

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}
    
    def __repr__(self):
        return '<Document ModelRelation ({}, {}, {})>'.format(
            self.uuid,
            self.name,
            self.project_id)


class MongoProject(Document):
    """
    Document of project in metadata backend storage.
    """
    
    uuid = SequenceField(db_alias=MONGO_DB_ALIAS_META_SERVICE)
    name = StringField(max_length=255, required=True, unique=True)
    properties = StringField(max_length=1000)
    uri = StringField(max_length=1000)
    is_deleted = BooleanField(default=False)

    model_relation = ListField(ReferenceField(MongoModelRelation))

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document Project ({}, {}, {}, {})>'.format(
            self.uuid,
            self.name,
            self.properties,
            self.uri)


class MongoProjectSnapshot(Document):
    """
    Document of project snapshot in metadata backend storage
    """

    uuid = SequenceField(db_alias=MONGO_DB_ALIAS_META_SERVICE)
    project_id = IntField()
    signature = StringField(max_length=255)
    create_time = LongField()

    model_version_relation = ListField(ReferenceField(MongoModelVersionRelation))

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document ProjectSnapshot ({}, {}, {}, {})>'.format(
            self.uuid,
            self.project_id,
            self.signature,
            self.create_time)


class MongoWorkflow(Document):
    """
    Document of workflow in metadata backend storage.
    """

    uuid = SequenceField(db_alias=MONGO_DB_ALIAS_META_SERVICE)
    name = StringField(max_length=255, required=True, unique=True)
    project_id = IntField()
    properties = StringField(max_length=1000)
    create_time = Column(BigInteger)
    update_time = Column(BigInteger)
    is_deleted = BooleanField(default=False)

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document Workflow ({}, {}, {}, {}, {}, {}, {})>'.format(
            self.uuid,
            self.name,
            self.project_id,
            self.properties,
            self.create_time,
            self.update_time,
            self.is_deleted)


class MongoModelVersion(Document):
    """
    Document of model version in Model Center backend storage.
    """

    model_name = StringField(max_length=255, required=True)
    model_version = StringField(max_length=10, required=True)
    model_path = StringField(max_length=500, default=None)
    model_type = StringField(max_length=500, default=None)
    version_desc = StringField(max_length=1000)
    version_status = StringField(max_length=20,
                                 default=ModelVersionStatus.to_string(ModelVersionStatus.READY))
    current_stage = StringField(max_length=20, default=STAGE_GENERATED)
    name_version_current_stage_unique = StringField(max_length=1000, required=True, unique=True)

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __init__(self, *args, **kwargs):
        n = kwargs['model_name']
        v = kwargs['model_version']
        c = kwargs['current_stage']
        name_version_current_stage = f'{n}-{v}-{c}'
        kwargs['name_version_current_stage_unique'] = name_version_current_stage
        super().__init__(*args, **kwargs)
    
    def __repr__(self):
        return '<Document ModelVersion ({}, {}, {}, {}, {}, {}, {}, {})>'.format(
            self.model_name,
            self.model_version,
            self.model_path,
            self.model_type,
            self.version_desc,
            self.version_status,
            self.current_stage)

    def to_meta_entity(self):
        return ModelVersionDetail(self.model_name,
                                  self.model_version,
                                  self.model_path,
                                  self.model_type,
                                  self.version_desc,
                                  self.version_status,
                                  self.current_stage)


class MongoRegisteredModel(Document):
    """
    Document of registered model in Model Center backend storage.
    """

    model_name = StringField(max_length=255, required=True, unique=True)
    model_desc = StringField(max_length=1000)

    model_version = ListField(ReferenceField(MongoModelVersion))

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document RegisteredModel ({}, {}, {})>'.format(
            self.model_name,
            self.model_desc)

    def to_meta_entity(self):
        return RegisteredModelDetail(self.model_name, self.model_desc)

    def to_detail_entity(self):
        latest_version = None
        for model_version in reversed(self.model_version):
            if model_version.current_stage != STAGE_DELETED:
                latest_version = model_version.to_meta_entity()
                break
        return RegisteredModelDetail(self.model_name, self.model_desc, latest_version)


class MongoArtifact(Document):
    """
    Document of artifact in metadata backend storage.
    """

    uuid = SequenceField(db_alias=MONGO_DB_ALIAS_META_SERVICE)
    name = StringField(max_length=255, required=True, unique=True)
    artifact_type = StringField(max_length=256)
    description = StringField(max_length=1000)
    uri = StringField(max_length=1000)
    create_time = LongField()
    update_time = LongField()
    properties = StringField(max_length=1000)
    is_deleted = BooleanField(default=False)

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document Artifact ({}, {}, {}, {}, {}, {}, {}, {}, {})>'.format(
            self.pk,
            self.name,
            self.artifact_type,
            self.description,
            self.uri,
            self.create_time,
            self.update_time,
            self.properties,
            self.is_deleted)


class MongoMetricMeta(Document):
    """
    Document of metric meta
    """

    uuid = SequenceField(db_alias=MONGO_DB_ALIAS_META_SERVICE)
    metric_name = StringField(max_length=256, required=True, unique=True)
    metric_type = StringField(max_length=256, default=MetricType.DATASET.value)
    metric_desc = StringField(max_length=4096)
    project_name = StringField(max_length=256, required=True)
    dataset_name = StringField(max_length=256)
    model_name = StringField(max_length=256)
    job_name = StringField(max_length=256)
    start_time = LongField()
    end_time = LongField()
    uri = StringField(max_length=1024)
    tags = StringField(max_length=256)
    properties = StringField(max_length=1024)
    is_deleted = BooleanField(default=False)

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document MetricMeta ({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})>' \
            .format(self.uuid, self.metric_name, self.metric_type, self.metric_desc, self.project_name,
                    self.dataset_name, self.model_name, self.start_time, self.end_time, self.uri, self.tags,
                    self.properties)


class MongoMetricSummary(Document):
    """
    Document of metric summary
    """

    uuid = SequenceField(db_alias=MONGO_DB_ALIAS_META_SERVICE)
    metric_name = StringField(max_length=256, required=True)
    metric_key = StringField(max_length=256, required=True)
    metric_value = StringField(max_length=2048, required=True)
    metric_timestamp = LongField()
    model_version = StringField(max_length=256)
    job_execution_id = StringField(max_length=256)
    is_deleted = BooleanField(default=False)

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document MetricSummary ({}, {}, {})>'.format(self.uuid, self.metric_name, self.metric_key,
                                                                        self.metric_value, self.metric_timestamp,
                                                                        self.model_version, self.job_execution_id)


class MongoMember(Document):
    """
    Document of cluster member.
    """

    version = LongField(required=True)
    server_uri = StringField(max_length=767, required=True, unique=True)
    update_time = LongField(required=True)
    uuid = StringField(max_length=128, required=True, unique=True)

    meta = {'db_alias': MONGO_DB_ALIAS_META_SERVICE}

    def __repr__(self):
        return '<Document Member ({}, {}, {}, {}, {})>'.format(
            self.id, self.version, self.server_uri, self.update_time, self.uuid)
